# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Chain constants"""

import os
from typing import cast

from aea.configurations.constants import CONTRACTS, PACKAGES
from aea.configurations.data_types import PublicId

from autonomy import AUTONOMY_DIR
from autonomy.constants import VALORY
from autonomy.data import DATA_DIR


CONTRACTS_DIR_FRAMEWORK = DATA_DIR / CONTRACTS
CONTRACTS_DIR_LOCAL = (
    AUTONOMY_DIR.parent / PACKAGES / VALORY / CONTRACTS
)  # use from an editable/local installation

ERC20_TOKEN_ADDRESS_LOCAL = "0x1291Be112d480055DaFd8a610b7d1e203891C274"  # nosec
DEFAULT_MULTISEND = "0x40A2aCCbd92BCA938b02010E17A5b8929b49130D"
CHAIN_PROFILES = {
    "local": {
        "component_registry": "0x5FbDB2315678afecb367f032d93F642f64180aa3",
        "agent_registry": "0xe7f1725E7734CE288F8367e1Bb143E90bb3F0512",
        "registries_manager": "0x9fE46736679d2D9a65F0992F2272dE9f3c7fa6e0",
        "service_registry": "0x998abeb3E57409262aE5b751f60747921B33613E",
        "service_manager": "0x4c5859f0F772848b2D91F1D83E2Fe57935348029",
        "gnosis_safe_proxy_factory": "0x0E801D84Fa97b50751Dbf25036d067dCf18858bF",
        "gnosis_safe_same_address_multisig": "0x8f86403A4DE0BB5791fa46B8e795C547942fE4Cf",
        "service_registry_token_utility": "0x36C02dA8a0983159322a80FFE9F24b1acfF8B570",
        "multisend": "0x9d4454B023096f34B160D6B654540c56A1F81688",
    },
    "ethereum": {
        "component_registry": "0x15bd56669F57192a97dF41A2aa8f4403e9491776",
        "agent_registry": "0x2F1f7D38e4772884b88f3eCd8B6b9faCdC319112",
        "registries_manager": "0x9eC9156dEF5C613B2a7D4c46C383F9B58DfcD6fE",
        "service_registry": "0x48b6af7B12C71f09e2fC8aF4855De4Ff54e775cA",
        "service_registry_token_utility": "0x3Fb926116D454b95c669B6Bf2E7c3bad8d19affA",
        "service_manager_token": "0x2EA682121f815FBcF86EA3F3CaFdd5d67F2dB143",
        "operator_whitelist": "0x42042799B0DE38AdD2a70dc996f69f98E1a85260",
        "gnosis_safe_proxy_factory": "0x46C0D07F55d4F9B5Eed2Fc9680B5953e5fd7b461",
        "gnosis_safe_same_address_multisig": "0xfa517d01DaA100cB1932FA4345F68874f7E7eF46",
        "staking_token": "0x0Dc23eEf3bC64CF3cbd8f9329B57AE4C4f28d5d2",
        "staking_verifier": "0x4503b79d468e81Ad2d4bd6Db991810da269bA777",
        "staking_factory": "0xEBdde456EA288b49f7D5975E7659bA1Ccf607efc",
    },
    "polygon": {
        "service_registry": "0xE3607b00E75f6405248323A9417ff6b39B244b50",
        "service_manager": "0x3C1fF68f5aa342D296d4DEe4Bb1cACCA912D95fE",
        "service_registry_token_utility": "0xa45E64d13A30a51b91ae0eb182e88a40e9b18eD8",
        "service_manager_token": "0x04b0007b2aFb398015B76e5f22993a1fddF83644",
        "gnosis_safe_proxy_factory": "0x3d77596beb0f130a4415df3D2D8232B3d3D31e44",
        "gnosis_safe_same_address_multisig": "0xd8BCC126ff31d2582018715d5291A508530587b0",
        "operator_whitelist": "0x526E064cB694E8f5B7DB299158e17F33055B3943",
        "staking_token": "0x4aba1Cf7a39a51D75cBa789f5f21cf4882162519",
        "staking_verifier": "0x8Bc0a5c5B1612A16983B01ecd7ae130E728390CB",
        "staking_factory": "0x46C0D07F55d4F9B5Eed2Fc9680B5953e5fd7b461",
    },
    "gnosis": {
        "service_registry": "0x9338b5153AE39BB89f50468E608eD9d764B755fD",
        "service_registry_token_utility": "0xa45E64d13A30a51b91ae0eb182e88a40e9b18eD8",
        "service_manager_token": "0x04b0007b2aFb398015B76e5f22993a1fddF83644",
        "operator_whitelist": "0x526E064cB694E8f5B7DB299158e17F33055B3943",
        "gnosis_safe_proxy_factory": "0x3C1fF68f5aa342D296d4DEe4Bb1cACCA912D95fE",
        "gnosis_safe_same_address_multisig": "0x6e7f594f680f7aBad18b7a63de50F0FeE47dfD06",
        "staking_token": "0xEa00be6690a871827fAfD705440D20dd75e67AB1",
        "staking_verifier": "0x1D59DadE4FAeA7771eC7221420012d413175404C",
        "staking_factory": "0xb0228CA253A88Bc8eb4ca70BCAC8f87b381f4700",
    },
    "arbitrum_one": {
        "service_registry": "0xE3607b00E75f6405248323A9417ff6b39B244b50",
        "service_registry_token_utility": "0x3d77596beb0f130a4415df3D2D8232B3d3D31e44",
        "service_manager_token": "0x34C895f302D0b5cf52ec0Edd3945321EB0f83dd5",
        "operator_whitelist": "0x3C1fF68f5aa342D296d4DEe4Bb1cACCA912D95fE",
        "gnosis_safe_proxy_factory": "0x63e66d7ad413C01A7b49C7FF4e3Bb765C4E4bd1b",
        "gnosis_safe_same_address_multisig": "0xBb7e1D6Cb6F243D6bdE81CE92a9f2aFF7Fbe7eac",
        "staking_token": "0x04b0007b2aFb398015B76e5f22993a1fddF83644",
        "staking_verifier": "0x7Fd1F4b764fA41d19fe3f63C85d12bf64d2bbf68",
        "staking_factory": "0xEB5638eefE289691EcE01943f768EDBF96258a80",
    },
    "optimistic": {
        "service_registry": "0x3d77596beb0f130a4415df3D2D8232B3d3D31e44",
        "service_registry_token_utility": "0xBb7e1D6Cb6F243D6bdE81CE92a9f2aFF7Fbe7eac",
        "service_manager_token": "0xFbBEc0C8b13B38a9aC0499694A69a10204c5E2aB",
        "operator_whitelist": "0x63e66d7ad413C01A7b49C7FF4e3Bb765C4E4bd1b",
        "gnosis_safe_proxy_factory": "0x5953f21495BD9aF1D78e87bb42AcCAA55C1e896C",
        "gnosis_safe_same_address_multisig": "0xb09CcF0Dbf0C178806Aaee28956c74bd66d21f73",
        "staking_token": "0x63C2c53c09dE534Dd3bc0b7771bf976070936bAC",
        "staking_verifier": "0x526E064cB694E8f5B7DB299158e17F33055B3943",
        "staking_factory": "0xa45E64d13A30a51b91ae0eb182e88a40e9b18eD8",
    },
    "base": {
        "service_registry": "0x3C1fF68f5aa342D296d4DEe4Bb1cACCA912D95fE",
        "service_registry_token_utility": "0x34C895f302D0b5cf52ec0Edd3945321EB0f83dd5",
        "service_manager_token": "0x63e66d7ad413C01A7b49C7FF4e3Bb765C4E4bd1b",
        "operator_whitelist": "0x3d77596beb0f130a4415df3D2D8232B3d3D31e44",
        "gnosis_safe_proxy_factory": "0x22bE6fDcd3e29851B29b512F714C328A00A96B83",
        "gnosis_safe_same_address_multisig": "0xFbBEc0C8b13B38a9aC0499694A69a10204c5E2aB",
        "staking_token": "0xEB5638eefE289691EcE01943f768EDBF96258a80",
        "staking_verifier": "0x10c5525F77F13b28f42c5626240c001c2D57CAd4",
        "staking_factory": "0x1cEe30D08943EB58EFF84DD1AB44a6ee6FEff63a",
    },
    "celo": {
        "service_registry": "0xE3607b00E75f6405248323A9417ff6b39B244b50",
        "service_registry_token_utility": "0x3d77596beb0f130a4415df3D2D8232B3d3D31e44",
        "service_manager_token": "0x34C895f302D0b5cf52ec0Edd3945321EB0f83dd5",
        "operator_whitelist": "0x3C1fF68f5aa342D296d4DEe4Bb1cACCA912D95fE",
        "gnosis_safe_proxy_factory": "0x63e66d7ad413C01A7b49C7FF4e3Bb765C4E4bd1b",
        "gnosis_safe_same_address_multisig": "0xBb7e1D6Cb6F243D6bdE81CE92a9f2aFF7Fbe7eac",
        "staking_token": "0xe1E1B286EbE95b39F785d8069f2248ae9C41b7a9",
        "staking_verifier": "0xc40C79C275F3fA1F3f4c723755C81ED2D53A8D81",
        "staking_factory": "0x1c2cD884127b080F940b7546c1e9aaf525b1FA55",
    },
    "mode": {
        "service_registry": "0x3C1fF68f5aa342D296d4DEe4Bb1cACCA912D95fE",
        "service_registry_token_utility": "0x34C895f302D0b5cf52ec0Edd3945321EB0f83dd5",
        "service_manager_token": "0x63e66d7ad413C01A7b49C7FF4e3Bb765C4E4bd1b",
        "operator_whitelist": "0x3d77596beb0f130a4415df3D2D8232B3d3D31e44",
        "gnosis_safe_proxy_factory": "0xBb7e1D6Cb6F243D6bdE81CE92a9f2aFF7Fbe7eac",
        "gnosis_safe_same_address_multisig": "0xFbBEc0C8b13B38a9aC0499694A69a10204c5E2aB",
        "staking_token": "0xE49CB081e8d96920C38aA7AB90cb0294ab4Bc8EA",
        "staking_native_token": "0x88DE734655184a09B70700aE4F72364d1ad23728",
        "staking_verifier": "0x87c511c8aE3fAF0063b3F3CF9C6ab96c4AA5C60c",
        "staking_factory": "0x75D529FAe220bC8db714F0202193726b46881B76",
    },
    "custom_chain": {
        "component_registry": cast(
            str, os.environ.get("CUSTOM_COMPONENT_REGISTRY_ADDRESS")
        ),
        "agent_registry": cast(str, os.environ.get("CUSTOM_AGENT_REGISTRY_ADDRESS")),
        "registries_manager": cast(
            str, os.environ.get("CUSTOM_REGISTRIES_MANAGER_ADDRESS")
        ),
        "service_manager": cast(str, os.environ.get("CUSTOM_SERVICE_MANAGER_ADDRESS")),
        "service_registry": cast(
            str, os.environ.get("CUSTOM_SERVICE_REGISTRY_ADDRESS")
        ),
        "gnosis_safe_proxy_factory": cast(
            str, os.environ.get("CUSTOM_GNOSIS_SAFE_PROXY_FACTORY_ADDRESS")
        ),
        "gnosis_safe_same_address_multisig": cast(
            str, os.environ.get("CUSTOM_GNOSIS_SAFE_SAME_ADDRESS_MULTISIG_ADDRESS")
        ),
        "service_registry_token_utility": cast(
            str, os.environ.get("CUSTOM_SERVICE_REGISTRY_TOKEN_UTILITY_ADDRESS")
        ),
        "multisend": cast(
            str,
            os.environ.get(
                "CUSTOM_MULTISEND_ADDRESS", "0x40A2aCCbd92BCA938b02010E17A5b8929b49130D"
            ),
        ),
    },
}


CHAIN_NAME_TO_CHAIN_ID = {
    "local": 31337,
    "ethereum": 1,
    "polygon": 137,
    "gnosis": 100,
    "arbitrum_one": 42161,
    "optimistic": 10,
    "base": 8453,
    "celo": 42220,
    "mode": 34443,
}

# Contract PublicIds
COMPONENT_REGISTRY_CONTRACT = PublicId.from_str("valory/component_registry")
AGENT_REGISTRY_CONTRACT = PublicId.from_str("valory/agent_registry")
REGISTRIES_MANAGER_CONTRACT = PublicId.from_str("valory/registries_manager")
SERVICE_MANAGER_CONTRACT = PublicId.from_str("valory/service_manager")
SERVICE_REGISTRY_CONTRACT = PublicId.from_str("valory/service_registry")
GNOSIS_SAFE_CONTRACT = PublicId.from_str("valory/gnosis_safe")
GNOSIS_SAFE_PROXY_FACTORY_CONTRACT = PublicId.from_str(
    "valory/gnosis_safe_proxy_factory"
)
GNOSIS_SAFE_SAME_ADDRESS_MULTISIG_CONTRACT = PublicId.from_str(
    "valory/gnosis_safe_same_address_multisig"
)
SERVICE_REGISTRY_TOKEN_UTILITY_CONTRACT = PublicId.from_str(
    "valory/service_registry_token_utility"
)
MULTISEND_CONTRACT = PublicId.from_str("valory/multisend")
ERC20_CONTRACT = PublicId.from_str("valory/erc20")

SERVICE_MANAGER_TOKEN_COMPATIBLE_CHAINS = (
    1,
    5,
    31337,
    100,
    10200,
    42161,
    421614,
    10,
    11155420,
    8453,
    84532,
    42220,
    42220,
)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for the subgraph client."""

from unittest import mock

from aea.configurations.data_types import PackageId, PackageType

from autonomy.chain.subgraph.client import SubgraphClient


class TestClient:
    """Test subgraph client"""

    client = SubgraphClient()

    def test_get_component_by_token(self) -> None:
        """Test get component by token query."""
        with mock.patch.object(SubgraphClient, "_query") as query:
            self.client.get_component_by_token(
                token_id=1,
                package_type=PackageType.PROTOCOL,
            )
            query.assert_called_once_with(
                '\nquery getUnit {\n  units(where:{tokenId: "1",'
                "packageType:protocol}){\n    tokenId\n    packageHash\n"
                "    publicId\n  }\n}\n"
            )

    def test_get_record_by_package_hash(self) -> None:
        """Test get record by package hash query."""
        with mock.patch.object(SubgraphClient, "_query") as query:
            self.client.get_record_by_package_hash(package_hash="bayfei")
            query.assert_called_once_with(
                '\nquery getUnit {\n  units(where:{packageHash:"bayfei"})'
                "{\n    tokenId\n    packageHash\n    publicId\n  }\n}\n"
            )

    def test_get_record_by_package_id(self) -> None:
        """Test get record by package hash query."""
        with mock.patch.object(SubgraphClient, "_query") as query:
            self.client.get_record_by_package_id(
                package_id=PackageId.from_uri_path("skill/valory/abci/0.1.0")
            )
            query.assert_called_once_with(
                '\nquery getUnit {\n  units(where:{publicId: "valory/abci"'
                ",packageType:skill}){\n    tokenId\n    packageHash\n    "
                "publicId\n  }\n}\n"
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests the behaviours of the slashing."""

# pylint: disable=protected-access

import json
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, Union, cast
from unittest import mock
from unittest.mock import MagicMock

import pytest

from packages.valory.contracts.gnosis_safe.contract import GnosisSafeContract
from packages.valory.contracts.service_registry.contract import ServiceRegistryContract
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.protocols.contract_api.custom_types import RawTransaction
from packages.valory.skills.abstract_round_abci.base import AbciAppDB, OffenceStatus
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    AsyncBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.abstract_round_abci.test_tools.base import (
    FSMBehaviourBaseCase,
)
from packages.valory.skills.abstract_round_abci.utils import inverse
from packages.valory.skills.slashing_abci.behaviours import (
    SlashingAbciBehaviours,
    SlashingCheckBehaviour,
    StatusResetBehaviour,
)
from packages.valory.skills.slashing_abci.rounds import Event, SynchronizedData
from packages.valory.skills.transaction_settlement_abci.behaviours import (
    TransactionSettlementRoundBehaviour,
)
from packages.valory.skills.transaction_settlement_abci.rounds import TX_HASH_LENGTH


STUB_OPERATORS_MAPPING = {
    "test_instance_1": "test_operator",
    "test_instance_0": "test_operator",
}
DUMMY_SLASH_THRESHOLD = 1
DUMMY_SLASH_COOLDOWN = 1


class BaseSlashingTest(FSMBehaviourBaseCase):
    """Base test case."""

    path_to_skill = Path(__file__).parents[1]
    behaviour: SlashingAbciBehaviours
    behaviour_class: Type[BaseBehaviour] = SlashingCheckBehaviour
    next_behaviour_class: Type[
        BaseBehaviour
    ] = TransactionSettlementRoundBehaviour.initial_behaviour_cls
    synchronized_data: SynchronizedData
    done_event = Event.SLASH_START

    @property
    def current_behaviour(
        self,
    ) -> Union[BaseBehaviour, SlashingCheckBehaviour, StatusResetBehaviour]:
        """Get the current behaviour."""
        return cast(
            Union[BaseBehaviour, SlashingCheckBehaviour, StatusResetBehaviour],
            self.behaviour.current_behaviour,
        )

    def fast_forward(self, data: Optional[Dict[str, Any]] = None) -> None:
        """Fast-forward on initialization"""

        data = data or {}
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.auto_behaviour_id(),
            SynchronizedData(AbciAppDB(setup_data=AbciAppDB.data_to_lists(data))),
        )
        assert (
            self.behaviour.current_behaviour is not None
            and self.behaviour.current_behaviour.behaviour_id
            == self.behaviour_class.auto_behaviour_id()
        )
        self.current_behaviour.params.__dict__["_frozen"] = False
        self.current_behaviour.params.slash_threshold_amount = DUMMY_SLASH_THRESHOLD
        self.current_behaviour.params.slash_cooldown_hours = DUMMY_SLASH_COOLDOWN
        self.current_behaviour.params.__dict__["_frozen"] = True

    def complete(self) -> None:
        """Complete test"""
        self.mock_a2a_transaction()
        self.end_round(done_event=self.done_event)
        assert (
            self.behaviour.current_behaviour is not None
            and self.behaviour.current_behaviour.behaviour_id
            == self.next_behaviour_class.auto_behaviour_id()
        )


class TestSlashingCheckBehaviour(BaseSlashingTest):
    """Tests for `SlashingCheckBehaviour`."""

    behaviour_class = SlashingCheckBehaviour

    def _mock_get_slash_data(
        self, error: bool = False, invalid_data: bool = False
    ) -> None:
        """Mock a ServiceRegistryContract.get_slash_data() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.RAW_TRANSACTION
            response_body = {} if invalid_data else {"data": b"slash_data"}
        else:
            response_performative = ContractApiMessage.Performative.ERROR
            response_body = {}

        self.mock_contract_api_request(
            contract_id=str(ServiceRegistryContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
                contract_address=self.current_behaviour.params.service_registry_address,
            ),
            response_kwargs=dict(
                performative=response_performative,
                raw_transaction=RawTransaction(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    def _mock_get_safe_tx_hash(
        self,
        error: bool = False,
        invalid_hash: bool = False,
    ) -> None:
        """Mock a GnosisSafeContract.get_raw_safe_transaction_hash() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.RAW_TRANSACTION
            response_body = (
                {"tx_hash": "0x"}
                if invalid_hash
                else {"tx_hash": "0x" + "0" * (TX_HASH_LENGTH - 2)}
            )
        else:
            response_performative = ContractApiMessage.Performative.ERROR
            response_body = {}

        self.mock_contract_api_request(
            contract_id=str(GnosisSafeContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
                contract_address=self.current_behaviour.synchronized_data.safe_contract_address,
            ),
            response_kwargs=dict(
                performative=response_performative,
                raw_transaction=RawTransaction(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    @mock.patch.object(AsyncBehaviour, "try_send")
    def test_get_callback_request(self, try_send_mock: mock.Mock) -> None:
        """Test `get_callback_request` method."""
        self.fast_forward(
            data=dict(
                slashing_in_flight=False,
                all_participants=["a"],
                participants=["a"],
                consensus_threshold=1,
            )
        )
        self.current_behaviour.round_sequence._offence_status = {"a": OffenceStatus()}
        self.current_behaviour.round_sequence._last_round_transition_timestamp = (
            datetime(2000, 1, 1)
        )

        # `__stopped` is `True` in the beginning
        callback = self.current_behaviour.get_callback_request()
        message_mock = MagicMock()
        callback(message_mock, MagicMock())
        try_send_mock.assert_not_called()

        # call act wrapper so that `__call_act_first_time()` is triggered, which sets `self.__stopped = False`
        # however, state is not `AsyncState.WAITING_MESSAGE`
        self.behaviour.act_wrapper()
        callback(message_mock, MagicMock())
        try_send_mock.assert_not_called()

        # call `wait_for_message()` which sets `self.__state = self.AsyncState.WAITING_MESSAGE`
        wait_gen = self.current_behaviour.wait_for_message()
        next(wait_gen)
        callback(message_mock, MagicMock())
        try_send_mock.assert_called_once_with(message_mock)

    @pytest.mark.parametrize(
        "offence_status, timestamps, last_timestamp, expected_amounts",
        (
            (
                {"agent": MagicMock(slash_amount=MagicMock(return_value=0))},
                {"agent": 0},
                datetime(2000, 1, 1),
                {},
            ),
            (
                {
                    "agent": MagicMock(
                        slash_amount=MagicMock(return_value=DUMMY_SLASH_THRESHOLD)
                    )
                },
                {"agent": 946684800 - DUMMY_SLASH_COOLDOWN},
                datetime(2000, 1, 1),
                {},
            ),
            (
                {
                    "agent": MagicMock(
                        slash_amount=MagicMock(return_value=DUMMY_SLASH_THRESHOLD + 1)
                    )
                },
                {"agent": 946684800},
                datetime(2000, 1, 1),
                {},
            ),
            (
                {
                    "agent": MagicMock(
                        slash_amount=MagicMock(return_value=DUMMY_SLASH_THRESHOLD + 1)
                    )
                },
                {"agent": 946684800 - DUMMY_SLASH_COOLDOWN - 1},
                datetime(2000, 1, 1),
                {"agent": DUMMY_SLASH_THRESHOLD + 1},
            ),
        ),
    )
    def test_check_offence_status(
        self,
        offence_status: Dict[str, MagicMock],
        timestamps: Dict[str, int],
        last_timestamp: datetime,
        expected_amounts: Dict[str, int],
    ) -> None:
        """Test the `_check_offence_status` method."""
        self.fast_forward(data={"slash_timestamps": json.dumps(timestamps)})
        # repeating this check for the `current_behaviour` here to avoid `mypy` reporting:
        # `error: Item "None" of "Optional[BaseBehaviour]" has no attribute "context"` when accessing the context below
        assert self.current_behaviour is not None

        self.current_behaviour._slash_amounts = {"agent": "something_random"}
        self.current_behaviour.round_sequence._offence_status = offence_status  # type: ignore
        self.current_behaviour.round_sequence._last_round_transition_timestamp = (
            last_timestamp
        )

        self.current_behaviour._check_offence_status()
        assert self.current_behaviour._slash_amounts == expected_amounts

    @dataclass
    class SlashingCheckBehaviourTestCase:
        """Test case parametrization for the `SlashingCheckBehaviour`."""

        slashing_in_flight: bool = False
        consensus_threshold: int = 1
        num_double_signed: int = 1
        slash_data_error: bool = False
        slash_data_invalid: bool = False
        safe_tx_error: bool = False
        safe_tx_invalid: bool = False

        @property
        def all_participants(self) -> List[str]:
            """Generate a stub list with participants."""
            return [f"participant{i}" for i in range(self.consensus_threshold)]

    @pytest.mark.parametrize(
        "test_case",
        (
            SlashingCheckBehaviourTestCase(slashing_in_flight=True),
            SlashingCheckBehaviourTestCase(consensus_threshold=2),
            SlashingCheckBehaviourTestCase(num_double_signed=0),
            SlashingCheckBehaviourTestCase(slash_data_error=True),
            SlashingCheckBehaviourTestCase(slash_data_invalid=True),
            SlashingCheckBehaviourTestCase(safe_tx_error=True),
            SlashingCheckBehaviourTestCase(safe_tx_invalid=True),
            SlashingCheckBehaviourTestCase(),
        ),
    )
    @mock.patch.object(AsyncBehaviour, "sleep")
    def test_slashing_check_act(
        self, sleep_mock: mock.Mock, test_case: SlashingCheckBehaviourTestCase
    ) -> None:
        """Test `PendingOffencesBehaviour`'s async act."""
        first_participant = test_case.all_participants[0]
        self.fast_forward(
            data=dict(
                slashing_in_flight=test_case.slashing_in_flight,
                all_participants=test_case.all_participants,
                participants=[first_participant],
                consensus_threshold=test_case.consensus_threshold,
                safe_contract_address="test_safe_contract_address",
            )
        )
        self.current_behaviour.round_sequence._offence_status = {
            first_participant: OffenceStatus()
        }
        self.current_behaviour.round_sequence._offence_status[
            first_participant
        ].num_double_signed += test_case.num_double_signed
        self.current_behaviour.round_sequence._last_round_transition_timestamp = (
            datetime(2000, 1, 1)
        )
        self.behaviour.act_wrapper()
        if test_case.slashing_in_flight or test_case.consensus_threshold == 2:
            with pytest.raises(AssertionError):
                self._mock_get_slash_data(
                    test_case.slash_data_error, test_case.slash_data_invalid
                )
            return

        if test_case.num_double_signed == 0:
            with pytest.raises(AssertionError):
                self._mock_get_slash_data(
                    test_case.slash_data_error, test_case.slash_data_invalid
                )
            sleep_mock.assert_called_once_with(self.current_behaviour.params.sleep_time)
            return

        self._mock_get_slash_data(
            test_case.slash_data_error, test_case.slash_data_invalid
        )

        if test_case.slash_data_error or test_case.slash_data_invalid:
            with pytest.raises(AssertionError):
                self._mock_get_safe_tx_hash(
                    test_case.safe_tx_error, test_case.safe_tx_invalid
                )
            return

        self._mock_get_safe_tx_hash(test_case.safe_tx_error, test_case.safe_tx_invalid)

        if test_case.safe_tx_error or test_case.safe_tx_invalid:
            with pytest.raises(AssertionError):
                self.complete()
            return

        self.complete()
        sleep_mock.assert_not_called()


class TestStatusResetBehaviour(BaseSlashingTest):
    """Tests for `StatusResetBehaviour`."""

    behaviour_class = StatusResetBehaviour
    next_behaviour_class = StatusResetBehaviour
    done_event = Event.SLASH_END

    def _mock_process_slash_receipt(
        self,
        error: bool = False,
    ) -> None:
        """Mock a ServiceRegistryContract.process_slash_receipt() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.RAW_TRANSACTION
            response_body = {
                "slash_timestamp": 10,
                "events": [{"amount": 2, "operator": "test_operator", "serviceId": 1}],
            }
        else:
            response_performative = ContractApiMessage.Performative.ERROR
            response_body = {}

        self.mock_contract_api_request(
            contract_id=str(ServiceRegistryContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
                contract_address=self.current_behaviour.params.service_registry_address,
            ),
            response_kwargs=dict(
                performative=response_performative,
                raw_transaction=RawTransaction(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    def _mock_get_instances_mapping(
        self,
        error: bool = False,
    ) -> None:
        """Mock a ServiceRegistryContract.get_operators_mapping() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.RAW_TRANSACTION
            response_body = STUB_OPERATORS_MAPPING
        else:
            response_performative = ContractApiMessage.Performative.ERROR
            response_body = {}

        self.mock_contract_api_request(
            contract_id=str(ServiceRegistryContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
                contract_address=self.current_behaviour.params.service_registry_address,
            ),
            response_kwargs=dict(
                performative=response_performative,
                raw_transaction=RawTransaction(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    @dataclass
    class StatusResetBehaviourTestCase:
        """Test case parametrization for the `StatusResetBehaviour`."""

        operators_mapping: Optional[str] = None
        get_mapping_error: bool = False
        process_receipt_error: bool = False

    @pytest.mark.parametrize(
        "test_case",
        (
            StatusResetBehaviourTestCase(get_mapping_error=True),
            StatusResetBehaviourTestCase(process_receipt_error=True),
            StatusResetBehaviourTestCase(
                operators_mapping=json.dumps(
                    inverse(dict(sorted(STUB_OPERATORS_MAPPING.items())))
                )
            ),
            StatusResetBehaviourTestCase(),
        ),
    )
    @mock.patch.object(AsyncBehaviour, "sleep")
    def test_status_reset_act(
        self, sleep_mock: mock.Mock, test_case: StatusResetBehaviourTestCase
    ) -> None:
        """Test `PendingOffencesBehaviour`'s async act."""
        self.fast_forward(
            data=dict(
                final_tx_hash="final_tx_hash",
                all_participants=["a"],
                participants=["a"],
                consensus_threshold=1,
                operators_mapping=test_case.operators_mapping,
            )
        )
        # initialize the offence status with a fake offence
        first_instance = list(STUB_OPERATORS_MAPPING.keys())[0]
        initial_offence_status = {
            instance: OffenceStatus() for instance in STUB_OPERATORS_MAPPING
        }
        offence_status = deepcopy(initial_offence_status)
        offence_status[first_instance].num_double_signed = 1
        self.current_behaviour.round_sequence._offence_status = deepcopy(offence_status)
        self.behaviour.act_wrapper()

        if test_case.operators_mapping is None and test_case.get_mapping_error:
            self._mock_get_instances_mapping(test_case.get_mapping_error)
            sleep_mock.assert_called_once_with(self.current_behaviour.params.sleep_time)
            with pytest.raises(AssertionError):
                self._mock_process_slash_receipt(test_case.process_receipt_error)
            return

        if test_case.operators_mapping is None:
            self._mock_get_instances_mapping(test_case.get_mapping_error)

        assert self.current_behaviour.offence_status == offence_status
        self._mock_process_slash_receipt(test_case.process_receipt_error)

        if test_case.process_receipt_error:
            sleep_mock.assert_called_once_with(self.current_behaviour.params.sleep_time)
            with pytest.raises(AssertionError):
                self.complete()
            return

        assert self.current_behaviour.offence_status == initial_offence_status
        # check the order for determinism
        assert list(self.current_behaviour.offence_status.keys()) == sorted(
            initial_offence_status.keys()
        )
        self.complete()
        sleep_mock.assert_not_called()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the models.py module of the skill."""

from unittest.mock import MagicMock

from hypothesis import given, settings
from hypothesis import strategies as st

from packages.valory.skills.abstract_round_abci.tests.conftest import (
    irrelevant_config,
    profile_name,
)
from packages.valory.skills.offend_abci.models import OffendParams, SharedState
from packages.valory.skills.offend_abci.rounds import Event, OffendAbciApp


settings.load_profile(profile_name)


class TestModels:
    """Test the models."""

    @staticmethod
    def test_setup() -> None:
        """Test `SharedState`'s `setup`."""
        shared_state = SharedState(name="", skill_context=MagicMock())
        shared_state.context.params.setup_params = {"test": []}
        shared_state.setup()
        assert (
            OffendAbciApp.event_to_timeout[Event.ROUND_TIMEOUT]
            == shared_state.context.params.round_timeout_seconds
        )

    @staticmethod
    @given(
        offend_config=st.fixed_dictionaries(
            dict(
                validator_downtime=st.booleans(),
                invalid_payload=st.booleans(),
                blacklisted=st.booleans(),
                suspected=st.booleans(),
                num_unknown=st.integers(),
                num_double_signed=st.integers(),
                num_light_client_attack=st.integers(),
            )
        )
    )
    def test_init(offend_config: dict) -> None:
        """Test the initialization of the `OffendParams`."""
        base_config = dict(
            name="test",
            skill_context=MagicMock(),
            service_id="test",
        )
        params = OffendParams(
            **base_config,
            **offend_config,
            **irrelevant_config,
        )
        assert all(getattr(params, name) == val for name, val in offend_config.items())



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2025 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the data classes for the Hello World ABCI application."""
import json
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, cast

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppTransitionFunction,
    AppState,
    BaseSynchronizedData,
    CollectSameUntilThresholdRound,
    DegenerateRound,
    get_name,
)
from packages.valory.skills.test_solana_tx_abci.payloads import SolanaTransactionPayload


class Event(Enum):
    """Event enumeration for the solana test skill."""

    DONE = "done"
    ROUND_TIMEOUT = "round_timeout"
    ERROR = "error"
    NO_MAJORITY = "no_majority"


class SynchronizedData(BaseSynchronizedData):
    """
    Class to represent the synchronized data.

    This data is replicated by the tendermint application.
    """

    @property
    def most_voted_instruction_set(self) -> List[Dict[str, Any]]:
        """Return the list of instructions selected for execution on the multisig."""
        return cast(
            List[Dict[str, Any]], self.db.get_strict("most_voted_instruction_set")
        )


class SolanaRound(CollectSameUntilThresholdRound):
    """Dummy solana round."""

    payload_class = SolanaTransactionPayload
    synchronized_data_class = BaseSynchronizedData
    extended_requirements = ()

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """End block."""
        if self.threshold_reached:  # pragma: no cover
            payload = cast(str, self.most_voted_payload)
            instructions = json.loads(payload)
            synchronized_data = self.synchronized_data.update(
                synchronized_data_class=SynchronizedData,
                **{
                    get_name(SynchronizedData.most_voted_instruction_set): instructions,
                }
            )

            return synchronized_data, Event.DONE
        if not self.is_majority_possible(
            self.collection, self.synchronized_data.nb_participants
        ):
            return self.synchronized_data, Event.NO_MAJORITY

        return None


class FinishedWithTransactionRound(DegenerateRound):
    """Finished with transaction round."""


class SolanaTestAbciApp(AbciApp[Event]):
    """SolanaTestAbciApp

    Initial round: SolanaRound

    Initial states: {SolanaRound}

    Transition states:
        0. SolanaRound
            - done: 1.
            - no majority: 0.
            - round timeout: 0.
        1. FinishedWithTransactionRound

    Final states: {FinishedWithTransactionRound}

    Timeouts:
        round timeout: 30.0
        error: 30.0
    """

    initial_round_cls: AppState = SolanaRound
    transition_function: AbciAppTransitionFunction = {
        SolanaRound: {
            Event.DONE: FinishedWithTransactionRound,
            Event.NO_MAJORITY: SolanaRound,
            Event.ROUND_TIMEOUT: SolanaRound,
        },
        FinishedWithTransactionRound: {},
    }
    final_states = {
        FinishedWithTransactionRound,
    }
    db_pre_conditions = {SolanaRound: set()}
    db_post_conditions = {
        FinishedWithTransactionRound: {
            get_name(SynchronizedData.most_voted_instruction_set)
        }
    }
    event_to_timeout: Dict[Event, float] = {
        Event.ROUND_TIMEOUT: 30.0,
        Event.ERROR: 30.0,
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test for contract module."""

from pathlib import Path

from aea_test_autonomy.base_test_classes.contracts import BaseRegistriesContractsTest
from aea_test_autonomy.docker.base import skip_docker_tests
from aea_test_autonomy.docker.registries import AGENT_REGISTRY

from packages.valory.contracts.agent_registry.contract import AgentRegistryContract


PACKAGE_DIR = Path(__file__).parent.parent


@skip_docker_tests
class TestAgentRegistry(BaseRegistriesContractsTest):
    """Test agent registry."""

    contract: AgentRegistryContract
    contract_address = AGENT_REGISTRY
    contract_directory = PACKAGE_DIR

    def test_get_token_uri(self) -> None:
        """Test get token URI method."""

        token_uri = self.contract.get_token_uri(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            token_id=1,
        )

        assert (
            token_uri
            == "https://gateway.autonolas.tech/ipfs/f01701220985b4c36158b51f8a865faceff8141dbc0989c349a1a41ba1e2ac8e5b24536b2"  # nosec
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Localhost Deployment utilities."""
import json
import os
import platform
import shutil
import subprocess  # nosec
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional

from aea.configurations.constants import AGENT

from autonomy.chain.config import ChainType
from autonomy.deploy.constants import (
    BENCHMARKS_DIR,
    TENDERMINT_BIN_UNIX,
    TENDERMINT_BIN_WINDOWS,
)


LOCAL_TENDERMINT_VERSION = "0.34.19"


def check_tendermint_version() -> Path:
    """Check tendermint version."""
    tendermint_executable = Path(str(shutil.which(TENDERMINT_BIN_UNIX)))
    if platform.system() == "Windows":
        tendermint_executable = (
            Path(os.path.dirname(sys.executable)) / TENDERMINT_BIN_WINDOWS
        )

    if (  # check tendermint version
        tendermint_executable is None
        or (
            current_version := subprocess.check_output(  # nosec
                [tendermint_executable, "version"]
            )
            .strip()
            .decode()
        )
        != LOCAL_TENDERMINT_VERSION
    ):
        raise FileNotFoundError(
            f"Please install tendermint version {LOCAL_TENDERMINT_VERSION} "
            f"or build and run via docker by using the --docker flag."
            + f"\nYour tendermint version is: {current_version}"
            if current_version
            else ""
        )

    return tendermint_executable


def _run_aea_cmd(
    args: List[str],
    cwd: Optional[Path] = None,
    stdout: Optional[int] = None,
    stderr: Optional[int] = subprocess.PIPE,
    ignore_error: Optional[str] = None,
    **kwargs: Any,
) -> None:
    """Run an aea command in a subprocess."""
    result = subprocess.run(  # pylint: disable=subprocess-run-check # nosec
        args=[sys.executable, "-m", "aea.cli", *args],
        cwd=cwd,
        stdout=stdout,
        stderr=stderr,
        **kwargs,
    )
    if result.returncode != 0:
        result_error = result.stderr.decode()
        if ignore_error and ignore_error not in result_error:
            raise RuntimeError(f"Error running: {args} @ {cwd}\n{result_error}")


def _prepare_agent_env(working_dir: Path) -> Dict[str, Any]:
    """Prepare agent env, add keys, run aea commands."""
    env = json.loads((working_dir / "agent.json").read_text(encoding="utf-8"))

    # TODO: Dynamic port allocation, backport to service builder
    env["CONNECTION_ABCI_CONFIG_HOST"] = "localhost"
    env["CONNECTION_ABCI_CONFIG_PORT"] = "26658"

    for var in env:
        # Fix tendermint connection params
        if var.endswith("MODELS_PARAMS_ARGS_TENDERMINT_COM_URL"):
            env[var] = "http://localhost:8080"

        if var.endswith("MODELS_PARAMS_ARGS_TENDERMINT_URL"):
            env[var] = "http://localhost:26657"

        if var.endswith("MODELS_PARAMS_ARGS_TENDERMINT_P2P_URL"):
            env[var] = "localhost:26656"

        if var.endswith("MODELS_BENCHMARK_TOOL_ARGS_LOG_DIR"):
            benchmarks_dir = working_dir / BENCHMARKS_DIR
            benchmarks_dir.mkdir(exist_ok=True, parents=True)
            env[var] = str(benchmarks_dir.resolve())

    (working_dir / "agent.json").write_text(
        json.dumps(env, indent=2),
        encoding="utf-8",
    )

    return env


def setup_agent(working_dir: Path, agent_dir: Path, keys_file: Path) -> None:
    """Setup locally deployed agent."""
    env = _prepare_agent_env(working_dir)

    _run_aea_cmd(
        ["fetch", env["AEA_AGENT"], "--alias", AGENT],
        cwd=working_dir,
    )

    # add private keys
    shutil.copy(keys_file, agent_dir)
    _run_aea_cmd(
        ["add-key", ChainType.ETHEREUM.value],
        cwd=agent_dir,
        ignore_error="already present",
    )
    _run_aea_cmd(["issue-certificates"], cwd=agent_dir)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Helpers for minting components"""

import time
from datetime import datetime
from math import ceil
from typing import Callable, Dict, List, Optional, Tuple, cast

from aea.configurations.data_types import PublicId
from aea.crypto.base import Crypto, LedgerApi

from autonomy.chain.base import UnitType, registry_contracts
from autonomy.chain.config import ChainType, ContractConfigs
from autonomy.chain.constants import (
    AGENT_REGISTRY_CONTRACT,
    COMPONENT_REGISTRY_CONTRACT,
    REGISTRIES_MANAGER_CONTRACT,
    SERVICE_MANAGER_CONTRACT,
    SERVICE_REGISTRY_CONTRACT,
)
from autonomy.chain.exceptions import ComponentMintFailed, InvalidMintParameter
from autonomy.chain.tx import TxSettler


try:
    from web3.exceptions import Web3Exception
except (ModuleNotFoundError, ImportError):
    Web3Exception = Exception


DEFAULT_NFT_IMAGE_HASH = "bafybeiggnad44tftcrenycru2qtyqnripfzitv5yume4szbkl33vfd4abm"
DEFAULT_TRANSACTION_WAIT_TIMEOUT = 60.0


def transact(
    ledger_api: LedgerApi,
    crypto: Crypto,
    tx: Dict,
    max_retries: int = 5,
    sleep: float = 5.0,
    timeout: Optional[float] = None,
) -> Dict:
    """Make a transaction and return a receipt"""
    retries = 0
    tx_receipt = None
    tx_signed = crypto.sign_transaction(transaction=tx)
    tx_digest = ledger_api.send_signed_transaction(tx_signed=tx_signed)
    deadline = datetime.now().timestamp() + (
        timeout or DEFAULT_TRANSACTION_WAIT_TIMEOUT
    )
    while (
        tx_receipt is None
        and retries < max_retries
        and deadline >= datetime.now().timestamp()
    ):
        try:
            return ledger_api.api.eth.get_transaction_receipt(tx_digest)
        except Web3Exception:  # pylint: disable=broad-except
            time.sleep(sleep)
    raise TimeoutError("Timed out when waiting for transaction to go through")


def sort_service_dependency_metadata(
    agent_ids: List[int],
    number_of_slots_per_agents: List[int],
    cost_of_bond_per_agent: List[int],
) -> Tuple[List[int], ...]:
    """Sort service dependencies and their respective parameters"""

    ids_sorted = []
    slots_sorted = []
    securities_sorted = []

    for idx, n_slots, b_cost in sorted(
        zip(agent_ids, number_of_slots_per_agents, cost_of_bond_per_agent),
        key=lambda x: x[0],
    ):
        ids_sorted.append(idx)
        slots_sorted.append(n_slots)
        securities_sorted.append(b_cost)

    return ids_sorted, slots_sorted, securities_sorted


def get_min_threshold(n: int) -> int:
    """Calculate minimum threshold required for N number of agents."""
    return ceil((n * 2 + 1) / 3)


class MintManager:
    """Mint helper."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        ledger_api: LedgerApi,
        crypto: Crypto,
        chain_type: ChainType,
        dry_run: bool = False,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        sleep: Optional[float] = None,
    ) -> None:
        """Initialize object."""
        self.crypto = crypto
        self.ledger_api = ledger_api
        self.chain_type = chain_type
        self.timeout = timeout
        self.retries = retries
        self.sleep = sleep
        self.dry_run = dry_run

    def _transact(
        self,
        method: Callable,
        kwargs: Dict,
        build_tx_ctr: str,
        event: str,
        process_receipt_ctr: PublicId,
    ) -> List[Dict]:
        """Auxiliary method for minting components."""
        tx_settler = TxSettler(
            ledger_api=self.ledger_api,
            crypto=self.crypto,
            chain_type=self.chain_type,
            timeout=self.timeout,
            retries=self.retries,
            sleep=self.sleep,
        )
        receipt = tx_settler.transact(
            method=method,
            contract=build_tx_ctr,
            kwargs=kwargs,
            dry_run=self.dry_run,
        )
        if self.dry_run:
            print("=== Dry run output ===")
            print("Method: " + str(method).split(" ")[2])
            print(
                f"Contract: {ContractConfigs.get(name=build_tx_ctr).contracts[self.chain_type]}"
            )
            print("Kwargs: ")
            for key, val in kwargs.items():
                print(f"    {key}: {val}")
            print("Transaction: ")
            for key, val in receipt.items():
                print(f"    {key}: {val}")
            return []
        events = tx_settler.process(
            event=event,
            receipt=receipt,
            contract=process_receipt_ctr,
        ).get("events")
        return cast(List[Dict], events)

    def validate_address(self, address: str) -> str:
        """Validate address string."""
        try:
            return self.ledger_api.api.to_checksum_address(address)
        except ValueError as e:  # pragma: nocover
            raise ComponentMintFailed(f"Invalid owner address {address}") from e

    def mint_component(
        self,
        metadata_hash: str,
        component_type: UnitType,
        owner: Optional[str] = None,
        dependencies: Optional[List[int]] = None,
    ) -> Optional[int]:
        """Publish component on-chain."""
        owner = self.validate_address(owner or self.crypto.address)
        if dependencies is not None:
            dependencies = sorted(set(dependencies))

        events = self._transact(
            method=registry_contracts.registries_manager.get_create_transaction,
            build_tx_ctr=REGISTRIES_MANAGER_CONTRACT.name,
            kwargs=dict(
                owner=owner,
                component_type=component_type,
                metadata_hash=metadata_hash,
                sender=self.crypto.address,
                dependencies=dependencies,
            ),
            event="CreateUnit",
            process_receipt_ctr=(
                COMPONENT_REGISTRY_CONTRACT
                if component_type == UnitType.COMPONENT
                else AGENT_REGISTRY_CONTRACT
            ),
        )
        for event in events:
            if (
                "unitHash" in event["args"]
                and event["args"]["unitHash"].hex() == metadata_hash[2:]
            ):
                return event["args"]["unitId"]
        return None

    def update_component(
        self, metadata_hash: str, unit_id: int, component_type: UnitType
    ) -> Optional[int]:
        """Update component on-chain."""

        events = self._transact(
            method=registry_contracts.registries_manager.get_update_hash_transaction,
            build_tx_ctr=REGISTRIES_MANAGER_CONTRACT.name,
            kwargs=dict(
                unit_id=unit_id,
                component_type=component_type,
                metadata_hash=metadata_hash,
                sender=self.crypto.address,
            ),
            event="UpdateUnitHash",
            process_receipt_ctr=(
                COMPONENT_REGISTRY_CONTRACT
                if component_type == UnitType.COMPONENT
                else AGENT_REGISTRY_CONTRACT
            ),
        )
        for event in events:
            if (
                "unitHash" in event["args"]
                and event["args"]["unitHash"].hex() == metadata_hash[2:]
            ):
                return event["args"]["unitId"]
        return None

    def mint_service(  # pylint: disable=too-many-arguments
        self,
        metadata_hash: str,
        agent_ids: List[int],
        number_of_slots_per_agent: List[int],
        cost_of_bond_per_agent: List[int],
        threshold: Optional[int] = None,
        token: Optional[str] = None,
        owner: Optional[str] = None,
    ) -> Optional[int]:
        """Publish component on-chain."""

        if len(agent_ids) == 0:
            raise InvalidMintParameter("Please provide at least one agent id")

        if len(number_of_slots_per_agent) == 0:
            raise InvalidMintParameter("Please for provide number of slots for agents")

        if len(cost_of_bond_per_agent) == 0:
            raise InvalidMintParameter("Please for provide cost of bond for agents")

        if (
            len(agent_ids) != len(number_of_slots_per_agent)
            or len(agent_ids) != len(cost_of_bond_per_agent)
            or len(number_of_slots_per_agent) != len(cost_of_bond_per_agent)
        ):
            raise InvalidMintParameter(
                "Make sure the number of agent ids, number of slots for agents and cost of bond for agents match"
            )

        if any(map(lambda x: x == 0, number_of_slots_per_agent)):
            raise InvalidMintParameter("Number of slots cannot be zero")

        if any(map(lambda x: x == 0, cost_of_bond_per_agent)):
            raise InvalidMintParameter("Cost of bond cannot be zero")

        number_of_agent_instances = sum(number_of_slots_per_agent)
        if threshold is None:
            threshold = get_min_threshold(number_of_agent_instances)

        if threshold < get_min_threshold(n=number_of_agent_instances):
            raise InvalidMintParameter(
                "The threshold value should at least be greater than or equal to ceil((n * 2 + 1) / 3), "
                "n is total number of agent instances in the service"
            )

        (
            agent_ids,
            number_of_slots_per_agent,
            cost_of_bond_per_agent,
        ) = sort_service_dependency_metadata(
            agent_ids=agent_ids,
            number_of_slots_per_agents=number_of_slots_per_agent,
            cost_of_bond_per_agent=cost_of_bond_per_agent,
        )

        owner = self.validate_address(owner or self.crypto.address)
        agent_params = [
            [n, c] for n, c in zip(number_of_slots_per_agent, cost_of_bond_per_agent)
        ]

        events = self._transact(
            method=registry_contracts.service_manager.get_create_transaction,
            build_tx_ctr=SERVICE_MANAGER_CONTRACT.name,
            kwargs=dict(
                metadata_hash=metadata_hash,
                agent_params=agent_params,
                agent_ids=agent_ids,
                threshold=threshold,
                token=token,
                owner=owner,
                sender=self.crypto.address,
            ),
            event="CreateService",
            process_receipt_ctr=SERVICE_REGISTRY_CONTRACT,
        )
        for event in events:
            if "serviceId" in event["args"]:
                return event["args"]["serviceId"]
        return None

    def update_service(  # pylint: disable=too-many-arguments
        self,
        metadata_hash: str,
        service_id: int,
        agent_ids: List[int],
        number_of_slots_per_agent: List[int],
        cost_of_bond_per_agent: List[int],
        threshold: Optional[int] = None,
        token: Optional[str] = None,
    ) -> Optional[int]:
        """Publish component on-chain."""

        if len(agent_ids) == 0:
            raise InvalidMintParameter("Please provide at least one agent id")

        if len(number_of_slots_per_agent) == 0:
            raise InvalidMintParameter("Please for provide number of slots for agents")

        if len(cost_of_bond_per_agent) == 0:
            raise InvalidMintParameter("Please for provide cost of bond for agents")

        if (
            len(agent_ids) != len(number_of_slots_per_agent)
            or len(agent_ids) != len(cost_of_bond_per_agent)
            or len(number_of_slots_per_agent) != len(cost_of_bond_per_agent)
        ):
            raise InvalidMintParameter(
                "Make sure the number of agent ids, number of slots for agents and cost of bond for agents match"
            )

        if any(map(lambda x: x == 0, number_of_slots_per_agent)):
            raise InvalidMintParameter("Number of slots cannot be zero")

        if any(map(lambda x: x == 0, cost_of_bond_per_agent)):
            raise InvalidMintParameter("Cost of bond cannot be zero")

        number_of_agent_instances = sum(number_of_slots_per_agent)
        if threshold is None:
            threshold = get_min_threshold(number_of_agent_instances)

        if threshold < get_min_threshold(number_of_agent_instances):
            raise InvalidMintParameter(
                "The threshold value should at least be greater than or equal to ceil((n * 2 + 1) / 3), "
                "n is total number of agent instances in the service"
            )

        (
            agent_ids,
            number_of_slots_per_agent,
            cost_of_bond_per_agent,
        ) = sort_service_dependency_metadata(
            agent_ids=agent_ids,
            number_of_slots_per_agents=number_of_slots_per_agent,
            cost_of_bond_per_agent=cost_of_bond_per_agent,
        )

        agent_params = [
            [n, c] for n, c in zip(number_of_slots_per_agent, cost_of_bond_per_agent)
        ]

        events = self._transact(
            method=registry_contracts.service_manager.get_update_transaction,
            build_tx_ctr=SERVICE_MANAGER_CONTRACT.name,
            kwargs=dict(
                service_id=service_id,
                sender=self.crypto.address,
                metadata_hash=metadata_hash,
                agent_ids=agent_ids,
                agent_params=agent_params,
                threshold=threshold,
                token=token,
            ),
            event="UpdateService",
            process_receipt_ctr=SERVICE_REGISTRY_CONTRACT,
        )
        for event in events:
            if (
                "configHash" in event["args"]
                and event["args"]["configHash"].hex() == metadata_hash[2:]
            ):
                return event["args"]["serviceId"]
        return None



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the scaffold contract definition."""

import json
from collections import OrderedDict
from enum import Enum, IntEnum
from typing import Any, Dict, List, Optional, Tuple, Union

from aea.common import JSONLike
from aea.configurations.base import PublicId
from aea.contracts.base import Contract
from aea.crypto.base import LedgerApi
from solders.system_program import transfer  # pylint: disable=import-error


Pubkey = Any  # defined in solders.pubkey
Keypair = Any  # defined in solders.keypair
Program = Any  # defined in anchorpy
Instruction = Any  # defined in solders.instruction


class SquadsMultisigAuthorityIndex(IntEnum):
    """Squads multisig authority index list."""

    INTERNAL = 0
    VAULT = 1
    SECONDARY = 2
    PROGRAM_UPGRADE = 3


class MultisigAccountType(Enum):
    """Squads multisig account types."""

    MS = "Ms"
    MS_TRANSACTION = "MsTransaction"
    MS_INSTRUCTION = "MsInstruction"

    def __str__(self) -> str:
        """String represenation."""
        return self._value_  # pylint: disable=no-member


class TransactionStatus(IntEnum):
    """Transaction status"""

    Draft = 0  # Transaction default state
    Active = 1  # Transaction is live and ready
    ExecuteReady = 2  # Transaction has been approved and is pending execution
    Executed = 3  # Transaction has been executed
    Rejected = 4  # Transaction has been rejected
    Cancelled = 5  # Transaction has been cancelled


def pubkey_container_to_list(container: Any) -> List[str]:
    """Convert pubkey container to list of pubkey strings."""
    return list(map(str, container))


def account_meta_container_to_list(container: Any) -> List[Dict[str, Any]]:
    """Convert pubkey container to list of pubkey strings."""
    return [
        {
            "pubkey": str(ac.pubkey),
            "is_signer": ac.is_signer,
            "is_writable": ac.is_writable,
        }
        for ac in container
    ]


def ms_account_to_dict(state: Any) -> Dict[str, Any]:
    """Deserialize multisig account state to a python dictionary."""
    return {
        "threshold": state.threshold,
        "authority_index": state.authority_index,
        "transaction_index": state.transaction_index,
        "ms_change_index": state.ms_change_index,
        "bump": state.bump,
        "create_key": str(state.create_key),
        "allow_external_execute": state.allow_external_execute,
        "keys": pubkey_container_to_list(container=state.keys),
    }


def ms_tx_account_to_dict(state: Any) -> Dict[str, Any]:
    """Deserialize multisig transaction account state to a python dictionary."""
    return {
        "creator": str(state.creator),
        "ms": str(state.ms),
        "transaction_index": state.transaction_index,
        "authority_index": state.authority_index,
        "authority_bump": state.authority_bump,
        "status": state.status.index,
        "instruction_index": state.instruction_index,
        "bump": state.bump,
        "approved": pubkey_container_to_list(state.approved),
        "rejected": pubkey_container_to_list(state.rejected),
        "cancelled": pubkey_container_to_list(state.cancelled),
        "executed_index": state.executed_index,
    }


def ms_ix_account_to_dict(state: Any) -> Dict[str, Any]:
    """Deserialize multisig instruction account to a python dictionary."""
    return {
        "program_id": str(state.program_id),
        "keys": account_meta_container_to_list(state.keys),
        "data": list(state.data),
        "instruction_index": state.instruction_index,
        "bump": state.bump,
        "executed": state.executed,
    }


ACCOUNT_STATE_SERIALIZERS = {
    MultisigAccountType.MS: ms_account_to_dict,
    MultisigAccountType.MS_TRANSACTION: ms_tx_account_to_dict,
    MultisigAccountType.MS_INSTRUCTION: ms_ix_account_to_dict,
}

# Same for mainnet, devnet, testnet
SQUADS_MULTISIG_ADDRESS = "SMPLecH534NA9acpos4G6x7uf3LWbCAwZQE9e8ZekMu"

# Note: When loading contract instances on the ethereum contract packages we use `contract_address`
# argument as the contract addres for the instance being loaded, but in case of squads multisig
# the main contract is defined a factory/proxy. Which means you cannot directly interact with a
# multisig, you'll have to go through the factory contract. So when loading the contract/program
# instance use the `SQUADS_MULTISIG_ADDRESS` (Check the `get_program_state` method for an exmaple).


class SquadsMultisig(Contract):
    """The scaffold contract class for a smart contract."""

    contract_id = PublicId.from_str("valory/squads_multisig:0.1.0")

    @classmethod
    def get_raw_transaction(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_RAW_TRANSACTION' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        raise NotImplementedError

    @classmethod
    def get_raw_message(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> bytes:
        """
        Handler method for the 'GET_RAW_MESSAGE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        raise NotImplementedError

    @classmethod
    def get_state(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_STATE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        raise NotImplementedError

    @classmethod
    def get_program_instance(cls, ledger_api: LedgerApi) -> Program:
        """
        Load multisig program instance.

        :param ledger_api: Ledger API instance.
        :return: Program instance for the squads multisig
        """
        program: Program = ledger_api.get_contract_instance(
            contract_interface=cls.contract_interface[ledger_api.identifier],
            contract_address=SQUADS_MULTISIG_ADDRESS,
        ).get("program")
        return program

    @classmethod
    def get_account_state(
        cls,
        ledger_api: LedgerApi,
        contract_address: Pubkey,
        account_type: Union[MultisigAccountType, str],
        program: Optional[Program] = None,
    ) -> JSONLike:
        """
        Get multisig account.

        Returns an object containing metadata for a squads multisig account.

        :param ledger_api: The ledger API object.
        :type ledger_api: LedgerApi
        :param program: The program instance.
        :type program: Program
        :param pda: The public key of the multisig account (Not the vault address).
        :type pda: Pubkey
        :return: The decoded multisig account information.
        :rtype: Any
        """
        account_type = MultisigAccountType(str(account_type))
        contract_address = ledger_api.to_pubkey(contract_address)
        account_info = ledger_api.api.get_account_info(pubkey=contract_address)
        program = program or cls.get_program_instance(ledger_api=ledger_api)
        state = program.account[  # pylint: disable=protected-access
            account_type.value
        ]._coder.accounts.decode(account_info.value.data)
        serializer = ACCOUNT_STATE_SERIALIZERS[account_type]
        return serializer(state=state)

    @classmethod
    def current_tx_index(
        cls,
        ledger_api: LedgerApi,
        contract_address: Pubkey,
        program: Optional[Program] = None,
    ) -> Dict[str, int]:
        """
        Get the current transaction index of a multisig.

        :param ledger_api: The ledger API.
        :type ledger_api: LedgerApi
        :param program: The program instance.
        :type program: Program
        :param multisig_pda: The public key of the multisig account (Not the vault address).
        :type multisig_pda: Pubkey
        :return: The transaction index for the account.
        :rtype: int
        """
        contract_address = ledger_api.to_pubkey(contract_address)
        program = program or cls.get_program_instance(ledger_api=ledger_api)
        account_data = cls.get_account_state(
            ledger_api=ledger_api,
            contract_address=contract_address,
            account_type=MultisigAccountType.MS,
            program=program,
        )
        return {"data": account_data["transaction_index"]}

    @classmethod
    def next_tx_index(
        cls,
        ledger_api: LedgerApi,
        contract_address: Pubkey,
        program: Optional[Program] = None,
    ) -> JSONLike:
        """
        Get the next transaction index of a multisig.

        :param ledger_api: The ledger API.
        :type ledger_api: LedgerApi
        :param program: The program instance.
        :type program: Program
        :param multisig_pda: The public key of the multisig account (Not the vault address).
        :type multisig_pda: Pubkey
        :return: The transaction index for the account.
        :rtype: int
        """
        index = 1 + cls.current_tx_index(
            ledger_api=ledger_api,
            contract_address=contract_address,
            program=program,
        ).pop("data")
        return dict(data=index)

    @classmethod
    def current_ix_index(
        cls,
        ledger_api: LedgerApi,
        contract_address: Pubkey,
        program: Optional[Program] = None,
    ) -> Dict[str, int]:
        """
        Get instruction index for a transaction.

        :param ledger_api: The ledger API.
        :type ledger_api: LedgerApi
        :param program: The program instance.
        :type program: Program
        :param tx_pda: The public key of the transaction account.
        :type tx_pda: Pubkey
        :return: The instruction index for the transaction.
        :rtype: int
        """
        contract_address = ledger_api.to_pubkey(contract_address)
        program = program or cls.get_program_instance(ledger_api=ledger_api)
        account_data = cls.get_account_state(
            ledger_api=ledger_api,
            contract_address=contract_address,
            account_type=MultisigAccountType.MS_TRANSACTION,
            program=program,
        )
        return {"data": account_data["instruction_index"]}

    @classmethod
    def next_ix_index(
        cls,
        ledger_api: LedgerApi,
        contract_address: Pubkey,
        program: Optional[Program] = None,
    ) -> Dict[str, int]:
        """
        Get the next transaction index of a multisig.

        :param ledger_api: The ledger API.
        :type ledger_api: LedgerApi
        :param program: The program instance.
        :type program: Program
        :param multisig_pda: The public key of the multisig account (Not the vault address).
        :type multisig_pda: Pubkey
        :return: The transaction index for the account.
        :rtype: int
        """
        next_ix = 1 + cls.current_ix_index(
            ledger_api=ledger_api,
            contract_address=contract_address,
            program=program,
        ).pop("data")
        return {"data": next_ix}

    @classmethod
    def get_tx_pda(
        cls,
        ledger_api: LedgerApi,
        contract_address: Pubkey,
        tx_index: int,
    ) -> JSONLike:
        """
        Get transaction PDA (Program Derived Address).

        :param ledger_api: The ledger API.
        :type ledger_api: LedgerApi
        :param contract_address: The address of the contract.
        :type contract_address: str
        :param tx_index: The index of the transaction.
        :type tx_index: int
        :param multisig_pda: The public key of the multisig account.
        :type multisig_pda: Pubkey
        :param program_id: The public key of the program.
        :type program_id: Pubkey
        :return: The transaction PDA and nonce.
        :rtype: Tuple[Pubkey, int]
        """
        contract_address = ledger_api.to_pubkey(contract_address)
        program_id = ledger_api.to_pubkey(SQUADS_MULTISIG_ADDRESS)
        seeds = [
            "squad".encode(encoding="utf-8"),
            bytes(contract_address),  # type: ignore
            int(str(tx_index), 10).to_bytes(byteorder="little", length=4),
            "transaction".encode(encoding="utf-8"),
        ]
        data = str(ledger_api.pda(seeds=seeds, program_id=program_id))
        return dict(data=data)

    @classmethod
    def get_ix_pda(
        cls,
        ledger_api: LedgerApi,
        contract_address: Pubkey,
        ix_index: int,
    ) -> Dict[str, Tuple[Pubkey, int]]:
        """Create TX PDA"""
        contract_address = ledger_api.to_pubkey(contract_address)
        program_id = ledger_api.to_pubkey(SQUADS_MULTISIG_ADDRESS)
        return {
            "data": ledger_api.pda(
                seeds=[
                    "squad".encode(encoding="utf-8"),
                    bytes(contract_address),  # type: ignore
                    int(str(ix_index), 10).to_bytes(byteorder="little", length=1),
                    "instruction".encode(encoding="utf-8"),
                ],
                program_id=program_id,
            )
        }

    @classmethod
    def create_transaction_ix(  # pylint: disable=too-many-arguments
        cls,
        ledger_api: LedgerApi,
        contract_address: Pubkey,
        authority_index: int,
        creator: Pubkey,
        tx_pda: Optional[Pubkey] = None,
    ) -> JSONLike:
        """
        Create instruction set for creating a transaction.

        :param ledger_api: The ledger API.
        :type ledger_api: LedgerApi
        :param contract_address: The address of the contract.
        :type contract_address: str
        :param program: The program instance.
        :type program: Program
        :param multisig_pda: The public key of the multisig account.
        :type multisig_pda: Pubkey
        :param authority_index: The index of the authority.
        :type authority_index: int
        :return: The created transaction tx.
        :rtype: JSONLike
        """
        creator = ledger_api.to_pubkey(creator)
        contract_address = ledger_api.to_pubkey(contract_address)
        program = cls.get_program_instance(ledger_api=ledger_api)
        if tx_pda is None:
            tx_pda = cls.get_tx_pda(
                ledger_api=ledger_api,
                contract_address=contract_address,
                tx_index=cls.next_tx_index(
                    ledger_api=ledger_api,
                    contract_address=contract_address,
                    program=program,
                ).pop("data"),
            )
        tx_pda = ledger_api.to_pubkey(tx_pda)
        ix = ledger_api.build_instruction(
            contract_instance=program,
            method_name="create_transaction",
            data=[authority_index],
            accounts={
                "multisig": contract_address,
                "transaction": tx_pda,
                "creator": creator,
                "system_program": ledger_api.system_program,
            },
            remaining_accounts=None,
        )
        return {"recent_blockhash": ledger_api.latest_hash, "ixs": [ix]}

    @classmethod
    def add_instruction_ix(  # pylint: disable=too-many-arguments
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        tx_pda: Pubkey,
        creator: Pubkey,
        ix: Instruction,
        index: Optional[int] = None,
    ) -> JSONLike:
        """Attach instruction"""
        tx_pda = ledger_api.to_pubkey(tx_pda)
        contract_address = ledger_api.to_pubkey(contract_address)
        creator = ledger_api.to_pubkey(creator)
        program = cls.get_program_instance(ledger_api=ledger_api)
        index = index or cls.next_ix_index(
            ledger_api=ledger_api,
            contract_address=contract_address,
            program=program,
        ).pop("data")
        ix_pda = cls.get_ix_pda(
            ledger_api=ledger_api,
            contract_address=tx_pda,
            ix_index=index,
        ).pop("data")
        ix = ledger_api.deserialize_ix(ix)
        incoming_is = program.type["IncomingInstruction"](
            program_id=ix.program_id,
            keys=[
                program.type["MsAccountMeta"](
                    pubkey=account.pubkey,
                    is_signer=account.is_signer,
                    is_writable=account.is_writable,
                )
                for account in ix.accounts
            ],
            data=ix.data,
        )
        ix = ledger_api.build_instruction(
            contract_instance=program,
            method_name="add_instruction",
            data=[incoming_is],
            accounts={
                "multisig": contract_address,
                "transaction": tx_pda,
                "instruction": ix_pda,
                "creator": creator,
                "system_program": ledger_api.system_program,
            },
        )
        return {"recent_blockhash": ledger_api.latest_hash, "ixs": [ix]}

    @classmethod
    def activate_transaction_ix(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        tx_pda: Pubkey,
        creator: Pubkey,
    ) -> JSONLike:
        """Create instruction set for activating a transaction."""
        contract_address = ledger_api.to_pubkey(contract_address)
        tx_pda = ledger_api.to_pubkey(tx_pda)
        creator = ledger_api.to_pubkey(creator)
        program = cls.get_program_instance(ledger_api=ledger_api)
        ix = ledger_api.build_instruction(
            contract_instance=program,
            method_name="activate_transaction",
            data=[],
            accounts={
                "multisig": contract_address,
                "transaction": tx_pda,
                "creator": creator,
            },
        )
        return {"recent_blockhash": ledger_api.latest_hash, "ixs": [ix]}

    @classmethod
    def create_new_transaction_ix(  # pylint: disable=too-many-arguments
        cls,
        ledger_api: LedgerApi,
        contract_address: Pubkey,
        authority_index: int,
        creator: Pubkey,
        ixs: List[Instruction],
        tx_pda: Optional[Pubkey] = None,
    ) -> JSONLike:
        """Create instruction set for creating and activating the transaction with given list of instruction."""
        creator = ledger_api.to_pubkey(creator)
        contract_address = ledger_api.to_pubkey(contract_address)
        program = cls.get_program_instance(ledger_api=ledger_api)
        if tx_pda is not None:
            tx_pda = ledger_api.to_pubkey(tx_pda)
        else:
            tx_pda = cls.get_tx_pda(
                ledger_api=ledger_api,
                contract_address=contract_address,
                tx_index=cls.next_tx_index(
                    ledger_api=ledger_api,
                    contract_address=contract_address,
                    program=program,
                ).pop("data"),
            )

        tx_ixs = []
        create_ix = cls.create_transaction_ix(
            ledger_api=ledger_api,
            contract_address=contract_address,
            authority_index=authority_index,
            creator=creator,
            tx_pda=tx_pda,
        )

        tx_ixs += create_ix["ixs"]
        for index, ix in enumerate(ixs):
            attach_ix = cls.add_instruction_ix(
                ledger_api=ledger_api,
                contract_address=contract_address,
                tx_pda=tx_pda,
                creator=creator,
                ix=ix,
                index=(index + 1),
            )
            tx_ixs += attach_ix["ixs"]

        activate_ix = cls.activate_transaction_ix(
            ledger_api=ledger_api,
            contract_address=contract_address,
            tx_pda=tx_pda,
            creator=creator,
        )
        tx_ixs += activate_ix["ixs"]
        return {"recent_blockhash": ledger_api.latest_hash, "ixs": tx_ixs}

    @classmethod
    def approve_transaction_ix(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        tx_pda: Pubkey,
        member: Pubkey,
    ) -> JSONLike:
        """Create instruction set for approving a transaction."""
        contract_address = ledger_api.to_pubkey(contract_address)
        tx_pda = ledger_api.to_pubkey(tx_pda)
        member = ledger_api.to_pubkey(member)
        program = cls.get_program_instance(ledger_api=ledger_api)
        ix = ledger_api.build_instruction(
            contract_instance=program,
            method_name="approve_transaction",
            data=[],
            accounts={
                "multisig": contract_address,
                "transaction": tx_pda,
                "member": member,
            },
        )
        return {"recent_blockhash": ledger_api.latest_hash, "ixs": [ix]}

    @classmethod
    def execute_transaction_ix(  # pylint: disable=too-many-arguments, too-many-locals
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        tx_pda: Pubkey,
        member: Pubkey,
    ) -> JSONLike:
        """Create instruction set for executing a transaction."""
        contract_address = ledger_api.to_pubkey(contract_address)
        tx_pda = ledger_api.to_pubkey(tx_pda)
        member = ledger_api.to_pubkey(member)
        program = cls.get_program_instance(ledger_api=ledger_api)
        ixs = []
        tx_account = cls.get_account_state(
            ledger_api=ledger_api,
            contract_address=tx_pda,
            account_type=MultisigAccountType.MS_TRANSACTION,
            program=program,
        )
        for idx in range(1, tx_account["instruction_index"] + 1):
            ix_pda = cls.get_ix_pda(
                ledger_api=ledger_api,
                contract_address=tx_pda,
                ix_index=idx,
            ).pop("data")
            ix_account = cls.get_account_state(
                ledger_api=ledger_api,
                contract_address=ix_pda,
                account_type=MultisigAccountType.MS_INSTRUCTION,
                program=program,
            )
            ixs.append({"pubkey": ix_pda, "account": ix_account})
        keys = OrderedDict()
        for ix in ixs:
            keys[ix["pubkey"]] = {
                "pubkey": ix["pubkey"],
                "is_signer": False,
                "is_writable": False,
            }
            keys[ledger_api.system_program] = {
                "pubkey": ledger_api.system_program,
                "is_signer": False,
                "is_writable": False,
            }
            for key in ix["account"]["keys"]:
                keys[key["pubkey"]] = {
                    "pubkey": key["pubkey"],
                    "is_signer": False,
                    "is_writable": key["is_writable"],
                }
        ix = ledger_api.build_instruction(
            contract_instance=program,
            method_name="execute_transaction",
            data=[bytes(list(range(len(keys))))],
            accounts={
                "multisig": contract_address,
                "transaction": tx_pda,
                "member": member,
            },
            remaining_accounts=[
                ledger_api.to_account_meta(**key) for key in keys.values()
            ],
        )
        return {"recent_blockhash": ledger_api.latest_hash, "ixs": [ix]}

    @classmethod
    def get_transfer_tx(  # pylint: disable=unused-argument
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        from_pubkey: str,
        to_pubkey: str,
        lamports: int,
    ) -> JSONLike:
        """Get transfer tx."""
        transfer_tx = transfer(
            params=dict(
                from_pubkey=ledger_api.to_pubkey(from_pubkey),
                to_pubkey=ledger_api.to_pubkey(to_pubkey),
                lamports=lamports,
            )
        ).to_json()
        return dict(data=json.loads(transfer_tx))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the handler for the offend_slash_abci."""

from packages.valory.skills.abstract_round_abci.handlers import (
    ABCIRoundHandler as BaseABCIRoundHandler,
)
from packages.valory.skills.abstract_round_abci.handlers import (
    ContractApiHandler as BaseContractApiHandler,
)
from packages.valory.skills.abstract_round_abci.handlers import (
    HttpHandler as BaseHttpHandler,
)
from packages.valory.skills.abstract_round_abci.handlers import (
    IpfsHandler as BaseIpfsHandler,
)
from packages.valory.skills.abstract_round_abci.handlers import (
    LedgerApiHandler as BaseLedgerApiHandler,
)
from packages.valory.skills.abstract_round_abci.handlers import (
    SigningHandler as BaseSigningHandler,
)
from packages.valory.skills.abstract_round_abci.handlers import (
    TendermintHandler as BaseTendermintHandler,
)


ABCIHandler = BaseABCIRoundHandler
HttpHandler = BaseHttpHandler
SigningHandler = BaseSigningHandler
LedgerApiHandler = BaseLedgerApiHandler
ContractApiHandler = BaseContractApiHandler
TendermintHandler = BaseTendermintHandler
IpfsHandler = BaseIpfsHandler



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test analyse ABCI docstrings."""

import difflib
import inspect
from pathlib import Path

from autonomy.analyse.abci.docstrings import (
    compare_docstring_content,
    docstring_abci_app,
)

from packages.valory.skills.offend_abci.rounds import OffendAbciApp


def test_docstring_abci_app() -> None:
    """Test docstring_abci_app"""

    expected = """\"\"\"OffendAbciApp

    Initial round: OffendRound

    Initial states: {OffendRound}

    Transition states:
        0. OffendRound
            - done: 1.
            - no majority: 0.
            - round timeout: 0.
        1. FinishedOffendRound

    Final states: {FinishedOffendRound}

    Timeouts:
        round timeout: 30.0
    \"\"\""""

    docstring = docstring_abci_app(OffendAbciApp)
    differences = "\n".join(difflib.unified_diff(docstring.split(), expected.split()))
    assert not differences, differences


def test_compare_docstring_content() -> None:
    """Test compare_docstring_content"""

    # no regex match
    assert compare_docstring_content("", "", "") == (False, "")

    # identical - no update
    docstring = docstring_abci_app(OffendAbciApp)
    abci_app_name = OffendAbciApp.__name__
    file_content = Path(inspect.getfile(OffendAbciApp)).read_text()
    result = compare_docstring_content(file_content, docstring, abci_app_name)
    assert result == (True, file_content)

    # mutated - update
    mutated_content = file_content.replace("Initial round: OffendRound", "")
    assert not mutated_content == file_content
    result = compare_docstring_content(mutated_content, docstring, abci_app_name)
    assert result == (True, file_content)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the composition for the test SOLANA skill."""
import packages.valory.skills.registration_abci.rounds as RegistrationAbci
import packages.valory.skills.reset_pause_abci.rounds as ResetAndPauseAbci
import packages.valory.skills.squads_transaction_settlement_abci.rounds as TransactionSubmissionAbciApp
import packages.valory.skills.test_solana_tx_abci.rounds as SolanaTestAbciApp
from packages.valory.skills.abstract_round_abci.abci_app_chain import (
    AbciAppTransitionMapping,
    chain,
)


# Here we define how the transition between the FSMs should happen
# more information here: https://docs.autonolas.network/fsm_app_introduction/#composition-of-fsm-apps
abci_app_transition_mapping: AbciAppTransitionMapping = {
    RegistrationAbci.FinishedRegistrationRound: SolanaTestAbciApp.SolanaRound,
    SolanaTestAbciApp.FinishedWithTransactionRound: TransactionSubmissionAbciApp.CreateTxRandomnessRound,
    TransactionSubmissionAbciApp.FinishedTransactionSubmissionRound: ResetAndPauseAbci.ResetAndPauseRound,
    ResetAndPauseAbci.FinishedResetAndPauseRound: SolanaTestAbciApp.SolanaRound,
    ResetAndPauseAbci.FinishedResetAndPauseErrorRound: RegistrationAbci.RegistrationRound,
}

ComposedAbciApp = chain(
    (
        RegistrationAbci.AgentRegistrationAbciApp,
        ResetAndPauseAbci.ResetPauseAbciApp,
        TransactionSubmissionAbciApp.SolanaTransactionSubmissionAbciApp,
        SolanaTestAbciApp.SolanaTestAbciApp,
    ),
    abci_app_transition_mapping,
)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test image build helpers."""

from aea.configurations.data_types import Dependency

from autonomy.deploy.image import generate_dependency_flag_var


def test_generate_dependency_flag_var() -> None:
    """Test generate_dependency_flag_var method."""

    dependencies = (
        Dependency(name="some-package", version="==1.0.0"),
        Dependency(
            name="other-package",
            git="https://github.com/author/some_package",
            ref="79342a93079648ef03ab5aaf14978068fc96587a",
        ),
    )

    assert generate_dependency_flag_var(dependencies=dependencies) == (
        "-e some-package==1.0.0 "
        "-e git+https://github.com/author/some_package@79342a93079648ef03ab5aaf14978068fc96587a#egg=other-package"
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test service helpers."""


from aea_test_autonomy.fixture_helpers import registries_scope_class  # noqa: F401

from autonomy.chain.service import get_reuse_multisig_payload

from tests.test_autonomy.test_chain.base import (
    AGENT_ID,
    BaseChainInteractionTest,
    COST_OF_BOND_FOR_AGENT,
    DUMMY_SERVICE,
    NUMBER_OF_SLOTS_PER_AGENT,
    THRESHOLD,
)


class TestReuseMultisigPayload(BaseChainInteractionTest):
    """Test get_reuse_multisig_payload method."""

    def test_inproper_termination(self) -> None:
        """Test inproper termination failure"""

        _, error = get_reuse_multisig_payload(
            ledger_api=self.ledger_api,
            crypto=self.crypto,
            chain_type=self.chain_type,
            service_id=1,
        )

        assert (
            error
            == "Service was not terminated properly, the service owner should be the only owner of the safe"
        )

    def test_no_previous_deployment(self) -> None:
        """Test inproper termination failure"""
        service_id = self.mint_component(
            package_id=DUMMY_SERVICE,
            service_mint_parameters=dict(
                agent_ids=[AGENT_ID],
                number_of_slots_per_agent=[NUMBER_OF_SLOTS_PER_AGENT],
                cost_of_bond_per_agent=[COST_OF_BOND_FOR_AGENT],
                threshold=THRESHOLD,
            ),
        )
        _, error = get_reuse_multisig_payload(
            ledger_api=self.ledger_api,
            crypto=self.crypto,
            chain_type=self.chain_type,
            service_id=service_id,
        )

        assert error == "Cannot reuse multisig, No previous deployment exist!"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/service_registry contract."""
from pathlib import Path
from typing import Any
from unittest import mock

import pytest
from aea_ledger_ethereum import EthereumCrypto
from aea_test_autonomy.base_test_classes.contracts import BaseRegistriesContractsTest
from aea_test_autonomy.docker.base import skip_docker_tests

from packages.valory.contracts.service_registry.contract import (
    DEPLOYED_BYTECODE_MD5_HASH_BY_CHAIN_ID,
    EXPECTED_CONTRACT_ADDRESS_BY_CHAIN_ID,
    ServiceRegistryContract,
)


PACKAGE_DIR = Path(__file__).parent.parent

SERVICE_REGISTRY_INVALID = "0x5aAeb6053F3E94C9b9A09f33669435E7Ef1BeAed"
VALID_SERVICE_ID = 1
INVALID_SERVICE_ID = 0
CHAIN_ID = 31337
AGENT_INSTANCES = [
    "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
    "0x70997970C51812dc3A010C7d01b50e0d17dc79C8",
    "0x3C44CdDdB6a900fa2b585dd299e03d12FA4293BC",
    "0x90F79bf6EB2c4f870365E785982E1f101E93b906",
]
OPERATOR = "0xBcd4042DE499D14e55001CcbB24a551F3b954096"
OPERATORS_MAPPING = dict.fromkeys(AGENT_INSTANCES, OPERATOR)


def event_filter_patch(event: str, return_value: Any) -> mock._patch:
    """Returns an event filter patch for the given event name."""
    return mock.patch.object(
        ServiceRegistryContract,
        "get_instance",
        return_value=mock.MagicMock(
            events=mock.MagicMock(
                **{
                    event: mock.MagicMock(
                        create_filter=lambda **_: mock.MagicMock(
                            get_all_entries=lambda *_: return_value
                        )
                    )
                }
            )
        ),
    )


class BaseServiceRegistryContractTest(BaseRegistriesContractsTest):
    """Base class for Service Registry contract tests"""

    contract: ServiceRegistryContract
    contract_address = EXPECTED_CONTRACT_ADDRESS_BY_CHAIN_ID[CHAIN_ID]
    invalid_contract_address = SERVICE_REGISTRY_INVALID
    path_to_contract = PACKAGE_DIR
    ledger_identifier = EthereumCrypto.identifier
    contract_directory = PACKAGE_DIR


@skip_docker_tests
class TestServiceRegistryContract(BaseServiceRegistryContractTest):
    """Test Service Registry Contract"""

    @pytest.mark.parametrize("valid_address", (True, False))
    def test_verify_contract(self, valid_address: bool) -> None:
        """Run verify test. If abi file is updated tests + addresses need updating"""
        bytecode = DEPLOYED_BYTECODE_MD5_HASH_BY_CHAIN_ID[CHAIN_ID]

        if valid_address:
            contract_address = self.contract_address
        else:
            contract_address = self.invalid_contract_address
            bytecode += "invalid"

        result = self.contract.verify_contract(
            self.ledger_api,
            contract_address,
        )

        assert result["verified"] is valid_address, result

    @pytest.mark.parametrize(
        "service_id, expected", [(INVALID_SERVICE_ID, False), (VALID_SERVICE_ID, True)]
    )
    def test_exists(self, service_id: int, expected: int) -> None:
        """Test whether service id exists"""
        exists = self.contract.exists(
            self.ledger_api,
            self.contract_address,
            service_id,
        )

        assert exists is expected

    def test_get_agent_instances(self) -> None:
        """Test agent instances retrieval"""

        return_value = {
            "numAgentInstances": 4,
            "agentInstances": AGENT_INSTANCES,
        }

        assert self.contract_address is not None

        result = self.contract.get_agent_instances(
            self.ledger_api,
            self.contract_address,
            VALID_SERVICE_ID,
        )

        assert result == return_value

    def test_get_service_owner(self) -> None:
        """Test service owner retrieval."""
        service_owner = AGENT_INSTANCES[0]
        assert self.contract_address is not None

        actual = self.contract.get_service_owner(
            self.ledger_api,
            self.contract_address,
            VALID_SERVICE_ID,
        )

        expected = dict(service_owner=service_owner)
        assert expected == actual

    def test_get_token_uri(self) -> None:
        """Test `get_token_uri` method."""

        token_uri = self.contract.get_token_uri(
            self.ledger_api,
            self.contract_address,
            VALID_SERVICE_ID,
        )

        assert (
            token_uri
            == "https://gateway.autonolas.tech/ipfs/f017012205555555555555555555555555555555555555555555555555555555555555555"  # nosec
        )

    def test_get_service_information(self) -> None:
        """Test `test_get_service_information` method."""

        (
            security_deposit,
            multisig_address,
            ipfs_hash_for_config,
            threshold,
            max_number_of_agent_instances,
            number_of_agent_instances,
            service_state,
            list_of_cannonical_agents,
        ) = self.contract.get_service_information(
            self.ledger_api,
            self.contract_address,
            VALID_SERVICE_ID,
        )

        assert security_deposit == 10000000000000000
        assert multisig_address == "0x77b783e911F4398D75908Cc60C7138Bd1eFe35Fd"
        assert ipfs_hash_for_config == b"UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU"
        assert threshold == 3
        assert max_number_of_agent_instances == 4
        assert number_of_agent_instances == 4
        assert service_state == 4
        assert list_of_cannonical_agents == [1]

    def test_get_slash_data(self) -> None:
        """Test the `get_slash_data`."""
        result = self.contract.get_slash_data(
            self.ledger_api,
            self.contract_address,
            AGENT_INSTANCES,
            [0, 0, 0, 1],
            service_id=1,
        )

        assert result.get("data", b"") == (
            b"s\xb8\xb6\xa2\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00`\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf3\x9f\xd6\xe5\x1a\xad\x88\xf6\xf4\xcej\xb8\x82ry\xcf"
            b'\xff\xb9"f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00p\x99yp\xc5\x18\x12\xdc:\x01\x0c}\x01\xb5\x0e\r'
            b"\x17\xdcy\xc8\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00<D\xcd\xdd\xb6\xa9\x00\xfa+X]\xd2\x99\xe0="
            b"\x12\xfaB\x93\xbc\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x90\xf7\x9b\xf6\xeb,O\x87\x03e\xe7\x85"
            b"\x98.\x1f\x10\x1e\x93\xb9\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"
        ), "Contract did not return the expected data."

    def test_get_operator(self) -> None:
        """Test `get_operator` method."""
        for agent_instance, expected_operator in OPERATORS_MAPPING.items():
            actual_operator = (
                self.contract._get_operator(  # pylint: disable=protected-access
                    self.ledger_api,
                    self.contract_address,
                    agent_instance,
                )
            )
            assert actual_operator == expected_operator

    def test_get_operators_mapping(self) -> None:
        """Test `get_operator` method."""
        actual_mapping = self.contract.get_operators_mapping(
            self.ledger_api,
            self.contract_address,
            frozenset(OPERATORS_MAPPING.keys()),
        )
        assert actual_mapping == OPERATORS_MAPPING



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests the slashing background rounds."""
from copy import deepcopy
from typing import cast
from unittest.mock import MagicMock

import pytest
from hypothesis import settings

from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    TransactionNotValidError,
)
from packages.valory.skills.abstract_round_abci.test_tools.rounds import (
    BaseRoundTestClass,
)
from packages.valory.skills.abstract_round_abci.tests.conftest import profile_name
from packages.valory.skills.slashing_abci.payloads import SlashingTxPayload
from packages.valory.skills.slashing_abci.rounds import Event, SlashingCheckRound
from packages.valory.skills.slashing_abci.rounds import (
    SynchronizedData as SlashingSyncedData,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    SynchronizedData as TxSettlementSyncedData,
)


settings.load_profile(profile_name)


class TestSlashingCheckRound(BaseRoundTestClass):
    """Tests for `SlashingCheckRound`."""

    _synchronized_data_class = SlashingSyncedData

    def test_run(self) -> None:
        """Run tests."""

        test_round = SlashingCheckRound(
            synchronized_data=self.synchronized_data,
            context=MagicMock(),
        )

        first_payload, *payloads = [
            SlashingTxPayload(
                sender,
                "tx_hex",
            )
            for sender in self.participants
        ]

        test_round.process_payload(first_payload)
        assert test_round.collection == {first_payload.sender: first_payload}
        res = test_round.end_block()
        assert res is None

        for payload in payloads[:2]:
            test_round.process_payload(payload)
            res = test_round.end_block()

        expected_state = cast(
            SlashingSyncedData,
            self.synchronized_data.update(
                termination_majority_reached=True,
            ),
        )

        assert res is not None
        actual_state, event = res
        actual_state = cast(SlashingSyncedData, actual_state)

        assert actual_state.slashing_in_flight == expected_state.slashing_in_flight
        assert (
            actual_state.slashing_majority_reached
            == expected_state.slashing_majority_reached
        )
        assert (
            TxSettlementSyncedData(actual_state.db).most_voted_tx_hash
            == TxSettlementSyncedData(expected_state.db).most_voted_tx_hash
        )
        assert event == Event.SLASH_START
        assert not test_round.collection

    def test_bad_payloads(self) -> None:
        """Tests the slashing check round when bad payloads are sent."""
        test_round = SlashingCheckRound(
            synchronized_data=deepcopy(self.synchronized_data),
            context=MagicMock(),
        )
        payload_data = "0xdata"
        bad_participant = "non_existent"
        bad_participant_payload = SlashingTxPayload(bad_participant, payload_data)
        first_payload, *_ = [
            SlashingTxPayload(participant, payload_data)
            for participant in self.participants
        ]

        with pytest.raises(
            TransactionNotValidError,
            match=f"{bad_participant} not in list of participants",
        ):
            test_round.check_payload(bad_participant_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match=f"{bad_participant} not in list of participants",
        ):
            test_round.process_payload(bad_participant_payload)

        # a valid payload gets sent for the first time and it goes through
        test_round.process_payload(first_payload)

        # a duplicate (valid) payload will not go through
        with pytest.raises(
            TransactionNotValidError,
            match=f"sender {first_payload.sender} has already sent value for round",
        ):
            test_round.check_payload(first_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match=f"sender {first_payload.sender} has already sent value for round",
        ):
            test_round.process_payload(first_payload)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Subgraph client."""

import os
from typing import List, Optional, cast

from aea.configurations.data_types import PackageId, PackageType
from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport
from typing_extensions import TypedDict

from autonomy.chain.subgraph.queries import (
    FIND_BY_PACKAGE_HASH,
    FIND_BY_PUBLIC_ID,
    FIND_BY_TOKEN_ID,
)


SUBGRAPH_URL = os.environ.get(
    "OPEN_AUTONOMY_SUBGRAPH_URL",
    "https://subgraph.autonolas.tech/subgraphs/name/autonolas",
)


class Unit(TypedDict):
    """Unit container."""

    tokenId: str
    packageHash: str
    publicId: str


class UnitContainer(TypedDict):
    """Unit container"""

    units: List[Unit]


class SubgraphClient:
    """Subgraph helper class."""

    client: Client

    def __init__(self, url: Optional[str] = None) -> None:
        """Initialize object"""

        self._url = url or SUBGRAPH_URL
        self._transport = RequestsHTTPTransport(
            url=self._url,
        )
        self.client = Client(
            transport=self._transport,
            fetch_schema_from_transport=True,
        )

    def _query(self, query: str) -> UnitContainer:
        """Perform a query"""
        return cast(UnitContainer, self.client.execute(gql(request_string=query)))

    def get_component_by_token(
        self, token_id: int, package_type: PackageType
    ) -> UnitContainer:
        """Get component by package hash"""
        return self._query(
            FIND_BY_TOKEN_ID.format(
                token_id=token_id,
                package_type=package_type.value,
            )
        )

    def get_record_by_package_hash(self, package_hash: str) -> UnitContainer:
        """Get component by package hash"""
        return self._query(
            FIND_BY_PACKAGE_HASH.format(
                package_hash=package_hash,
            )
        )

    def get_record_by_package_id(self, package_id: PackageId) -> UnitContainer:
        """Get component by package hash"""
        public_id = f"{package_id.author}/{package_id.name}"
        return self._query(
            FIND_BY_PUBLIC_ID.format(
                public_id=public_id,
                package_type=package_id.package_type.value,
            )
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the payloads of the slashing."""

import pytest

from packages.valory.skills.abstract_round_abci.base import ROUND_COUNT_DEFAULT
from packages.valory.skills.slashing_abci.payloads import (
    SlashingTxPayload,
    StatusResetPayload,
)


@pytest.mark.parametrize(
    "sender, tx_hex",
    (
        (
            "sender",
            "tx_hex",
        ),
    ),
)
def test_slashing_tx_payload(
    sender: str,
    tx_hex: str,
) -> None:
    """Test `SlashingTxPayload`."""

    payload = SlashingTxPayload(sender, tx_hex)

    assert payload.id_
    assert payload.round_count == ROUND_COUNT_DEFAULT
    assert payload.sender == sender
    assert payload.tx_hex == tx_hex
    assert payload.data == {
        "tx_hex": tx_hex,
        "in_progress": True,
        "sent": True,
    }


@pytest.mark.parametrize(
    "sender, operators_mapping, slash_timestamps",
    (
        (
            "sender",
            "mapping",
            "timestamps",
        ),
    ),
)
def test_status_reset_payload(
    sender: str,
    operators_mapping: str,
    slash_timestamps: str,
) -> None:
    """Test `StatusResetPayload`"""

    payload = StatusResetPayload(
        sender,
        operators_mapping,
        slash_timestamps,
    )

    assert payload.id_
    assert payload.round_count == ROUND_COUNT_DEFAULT
    assert payload.sender == sender
    assert payload.operators_mapping == operators_mapping
    assert payload.slash_timestamps == slash_timestamps
    assert payload.data == {
        "operators_mapping": operators_mapping,
        "slash_timestamps": slash_timestamps,
        "slashing_in_flight": False,
        "slash_tx_sent": False,
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Deploy CLI module."""

import os
import shutil
from pathlib import Path
from typing import List, Optional, cast

import click
from aea import AEA_DIR
from aea.cli.registry.settings import REGISTRY_REMOTE
from aea.cli.utils.click_utils import (
    password_option,
    registry_flag,
    reraise_as_click_exception,
)
from aea.cli.utils.context import Context
from aea.configurations.constants import DEFAULT_ENV_DOTFILE

from autonomy.chain.config import ChainType
from autonomy.cli.helpers.deployment import (
    build_and_deploy_from_token,
    build_deployment,
    run_deployment,
    run_host_deployment,
    stop_deployment,
)
from autonomy.cli.helpers.env import load_env_file
from autonomy.cli.utils.click_utils import (
    PathArgument,
    chain_selection_flag,
    image_author_option,
)
from autonomy.constants import DEFAULT_BUILD_FOLDER, DEFAULT_KEYS_FILE
from autonomy.deploy.base import (
    DEFAULT_AGENT_CPU_LIMIT,
    DEFAULT_AGENT_CPU_REQUEST,
    DEFAULT_AGENT_MEMORY_LIMIT,
    DEFAULT_AGENT_MEMORY_REQUEST,
    NotValidKeysFile,
    build_hash_id,
)
from autonomy.deploy.constants import INFO, LOGGING_LEVELS
from autonomy.deploy.generators.docker_compose.base import DockerComposeGenerator
from autonomy.deploy.generators.kubernetes.base import KubernetesGenerator
from autonomy.deploy.generators.localhost.base import HostDeploymentGenerator


def _validate_packages_path(path: Optional[Path] = None) -> Path:
    """Find packages dir for development mode."""
    path = Path(path or Path.cwd()).resolve()
    while path != path.root:
        if path.name == "packages" and (path / "packages.json").exists():
            return path
        path = path.parent
    raise click.ClickException(
        "Please provide path to packages directory for development mode"
    )


def _validate_open_aea_dir(path: Optional[Path] = None) -> Path:
    """Validate open-aea directory."""
    path = Path(path or Path(AEA_DIR).parent).resolve()
    for child in ("aea", "plugins", "setup.py"):
        if not (path / child).exists():
            raise click.ClickException(
                "Please provide valid open-aea repository path for development mode"
            )
    return path


@click.group(name="deploy")
@click.option(
    "--env-file",
    type=PathArgument(
        exists=True,
        dir_okay=False,
        file_okay=True,
    ),
    help="File containing environment variable mappings",
)
@click.pass_context
def deploy_group(
    click_context: click.Context,  # pylint: disable=unused-argument
    env_file: Optional[Path],
) -> None:
    """Deploy an agent service."""
    dot_env_file = Path.cwd() / DEFAULT_ENV_DOTFILE
    if dot_env_file.exists():
        load_env_file(file=dot_env_file)
    if env_file is not None:
        load_env_file(file=env_file, serialize_json=True)


@deploy_group.command(name="build")
@click.argument("keys_file", type=str, required=False)
@click.option(
    "--o",
    "output_dir",
    type=click.Path(exists=False, dir_okay=True),
    help="Path to output dir.",
)
@click.option(
    "--n",
    "number_of_agents",
    type=int,
    default=None,
    help="Number of agents.",
)
@click.option(
    "--number-of-services",
    "number_of_services",
    type=int,
    default=1,
    help="Number of services.",
)
@click.option(
    "--localhost",
    "deployment_type",
    flag_value=HostDeploymentGenerator.deployment_type,
    help="Use localhost as a backend.",
)
@click.option(
    "--docker",
    "deployment_type",
    flag_value=DockerComposeGenerator.deployment_type,
    default=True,
    help="Use docker as a backend. (default)",
)
@click.option(
    "--kubernetes",
    "deployment_type",
    flag_value=KubernetesGenerator.deployment_type,
    help="Use kubernetes as a backend.",
)
@click.option(
    "--dev",
    "dev_mode",
    is_flag=True,
    default=False,
    help="Create development environment.",
)
@click.option(
    "--log-level",
    type=click.Choice(choices=LOGGING_LEVELS, case_sensitive=True),
    help="Logging level for runtime.",
    default=INFO,
)
@click.option(
    "--packages-dir", type=click.Path(), help="Path to packages dir (Use with dev mode)"
)
@click.option(
    "--open-aea-dir",
    type=click.Path(),
    help="Path to open-aea repo (Use with dev mode)",
)
@click.option(
    "--aev",
    is_flag=True,
    default=False,
    help="Apply environment variable when loading service config.",
)
@click.option(
    "--use-hardhat",
    is_flag=True,
    default=False,
    help="Include a hardhat node in the deployment setup.",
)
@click.option(
    "--use-acn",
    is_flag=True,
    default=False,
    help="Include an ACN node in the deployment setup.",
)
@click.option(
    "-ltm",
    "--local-tm-setup",
    "use_tm_testnet_setup",
    is_flag=True,
    default=False,
    help="Use local tendermint chain setup.",
)
@click.option("--image-version", type=str, help="Define runtime image version.")
@click.option(
    "--agent-cpu-request",
    type=float,
    help="Set agent CPU usage request.",
    default=DEFAULT_AGENT_CPU_REQUEST,
)
@click.option(
    "--agent-memory-request",
    type=int,
    help="Set agent memory usage request.",
    default=DEFAULT_AGENT_MEMORY_REQUEST,
)
@click.option(
    "--agent-cpu-limit",
    type=float,
    help="Set agent CPU usage limit.",
    default=DEFAULT_AGENT_CPU_LIMIT,
)
@click.option(
    "--agent-memory-limit",
    type=int,
    help="Set agent memory usage limit.",
    default=DEFAULT_AGENT_MEMORY_LIMIT,
)
@click.option(
    "--mkdir",
    type=str,
    help="Directory names to create in the build directory.",
    default=[],
    multiple=True,
)
@registry_flag()
@password_option(confirmation_prompt=True)
@image_author_option
@click.pass_context
def build_deployment_command(  # pylint: disable=too-many-arguments, too-many-locals
    click_context: click.Context,
    keys_file: Optional[Path],
    deployment_type: str,
    output_dir: Optional[Path],
    dev_mode: bool,
    registry: str,
    mkdir: List[str],
    number_of_agents: Optional[int] = None,
    number_of_services: int = 1,
    password: Optional[str] = None,
    open_aea_dir: Optional[Path] = None,
    packages_dir: Optional[Path] = None,
    log_level: str = INFO,
    aev: bool = False,
    image_version: Optional[str] = None,
    use_hardhat: bool = False,
    use_acn: bool = False,
    use_tm_testnet_setup: bool = False,
    image_author: Optional[str] = None,
    agent_cpu_limit: Optional[float] = None,
    agent_memory_limit: Optional[int] = None,
    agent_cpu_request: Optional[float] = None,
    agent_memory_request: Optional[int] = None,
) -> None:
    """Build deployment setup for n agents."""
    if password is not None:  # pragma: nocover
        click.echo(
            "WARNING: `--password` flag has been deprecated, "
            "use `OPEN_AUTONOMY_PRIVATE_KEY_PASSWORD` to export the password value"
        )

    keys_file = Path(keys_file or DEFAULT_KEYS_FILE).absolute()
    if not keys_file.exists() and not dev_mode:
        message = f"No such file or directory: {keys_file}. Please provide valid path for keys file."
        raise click.ClickException(message)

    if dev_mode:
        packages_dir = _validate_packages_path(path=packages_dir)
        open_aea_dir = _validate_open_aea_dir(path=open_aea_dir)

    ctx = cast(Context, click_context.obj)
    ctx.registry_type = registry

    abci_build_count = len(
        [name for name in os.listdir() if name.startswith("abci_build_")]
    )

    for service_index in range(number_of_services):
        service_hash_id = build_hash_id()
        build_dir = Path(
            output_dir or DEFAULT_BUILD_FOLDER.format(service_hash_id)
        ).absolute()

        try:
            build_deployment(
                service_hash_id=service_hash_id,
                service_offset=abci_build_count + service_index,
                keys_file=keys_file,
                build_dir=build_dir,
                deployment_type=deployment_type,
                dev_mode=dev_mode,
                number_of_agents=number_of_agents,
                packages_dir=packages_dir,
                open_aea_dir=open_aea_dir,
                log_level=log_level,
                apply_environment_variables=aev,
                image_version=image_version,
                use_hardhat=use_hardhat,
                use_acn=use_acn,
                use_tm_testnet_setup=use_tm_testnet_setup,
                image_author=image_author,
                mkdir=mkdir,
                resources={
                    "agent": {
                        "limit": {"cpu": agent_cpu_limit, "memory": agent_memory_limit},
                        "requested": {
                            "cpu": agent_cpu_request,
                            "memory": agent_memory_request,
                        },
                    }
                },
            )
        except (NotValidKeysFile, FileNotFoundError, FileExistsError) as e:
            shutil.rmtree(build_dir)
            raise click.ClickException(str(e)) from e


@deploy_group.command(name="run")
@click.option(
    "--build-dir",
    type=click.Path(),
    help="Path to the deployment build directory.",
)
@click.option(
    "--no-recreate",
    is_flag=True,
    default=False,
    help="If containers already exist, don't recreate them.",
)
@click.option(
    "--remove-orphans",
    is_flag=True,
    default=False,
    help="Remove containers for services not defined in the Compose file.",
)
@click.option(
    "--detach",
    is_flag=True,
    default=False,
    help="Run service in the background.",
)
@click.option(
    "--localhost",
    "deployment_type",
    flag_value="localhost",
    help="Use localhost as a backend.",
)
@click.option(
    "--docker",
    "deployment_type",
    flag_value="docker",
    help="Use docker as a backend. (default)",
    default=True,
)
def run(
    build_dir: Path,
    no_recreate: bool,
    remove_orphans: bool,
    detach: bool,
    deployment_type: str,
) -> None:
    """Run deployment."""
    build_dir = Path(build_dir or Path.cwd()).absolute()
    deployment = (
        HostDeploymentGenerator
        if deployment_type == "localhost"
        else DockerComposeGenerator
    )
    if not (build_dir / deployment.output_name).exists():
        raise click.ClickException(
            f"Deployment configuration does not exist @ {build_dir}"
        )
    click.echo(f"Running build @ {build_dir}")
    if deployment_type == "localhost":
        run_host_deployment(build_dir, detach)
    else:
        run_deployment(build_dir, no_recreate, remove_orphans, detach=detach)


@deploy_group.command(name="stop")
@click.option(
    "--build-dir",
    type=click.Path(),
    help="Path to the deployment build directory.",
)
def stop(build_dir: Path) -> None:
    """Stop a running deployment."""
    build_dir = Path(build_dir or Path.cwd()).absolute()
    if not (build_dir / DockerComposeGenerator.output_name).exists():
        raise click.ClickException(
            f"Deployment configuration does not exist @ {build_dir}"
        )
    stop_deployment(build_dir=build_dir)


@deploy_group.command(name="from-token")
@click.argument("token_id", type=int)
@click.argument("keys_file", type=click.Path())
@click.option("--n", type=int, help="Number of agents to include in the build.")
@click.option("--skip-image", is_flag=True, default=False, help="Skip building images.")
@click.option(
    "--aev",
    is_flag=True,
    default=False,
    help="Apply environment variable when loading service config.",
)
@click.option(
    "--docker",
    "deployment_type",
    flag_value=DockerComposeGenerator.deployment_type,
    default=True,
    help="Use docker as a backend.",
)
@click.option(
    "--kubernetes",
    "deployment_type",
    flag_value=KubernetesGenerator.deployment_type,
    help="Use kubernetes as a backend.",
)
@click.option(
    "--no-deploy",
    is_flag=True,
    help="If set to true, the deployment won't run automatically",
)
@click.option(
    "--detach",
    is_flag=True,
    default=False,
    help="Run service in the background.",
)
@click.option(
    "--agent-cpu-request",
    type=float,
    help="Set agent CPU usage request.",
    default=DEFAULT_AGENT_CPU_REQUEST,
)
@click.option(
    "--agent-memory-request",
    type=int,
    help="Set agent memory usage request.",
    default=DEFAULT_AGENT_MEMORY_REQUEST,
)
@click.option(
    "--agent-cpu-limit",
    type=float,
    help="Set agent CPU usage limit.",
    default=DEFAULT_AGENT_CPU_LIMIT,
)
@click.option(
    "--agent-memory-limit",
    type=int,
    help="Set agent memory usage limit.",
    default=DEFAULT_AGENT_MEMORY_LIMIT,
)
@chain_selection_flag(help_string_format="Use {} chain to resolve the token id.")
@click.pass_context
@password_option(confirmation_prompt=True)
def run_deployment_from_token(  # pylint: disable=too-many-arguments, too-many-locals
    click_context: click.Context,
    token_id: int,
    keys_file: Path,
    chain_type: ChainType,
    skip_image: bool,
    n: Optional[int],
    deployment_type: str,
    no_deploy: bool,
    detach: bool,
    aev: bool = False,
    password: Optional[str] = None,
    agent_cpu_limit: Optional[float] = None,
    agent_memory_limit: Optional[int] = None,
    agent_cpu_request: Optional[float] = None,
    agent_memory_request: Optional[int] = None,
) -> None:
    """Run service deployment."""
    if password is not None:  # pragma: nocover
        click.echo(
            "WARNING: `--password` flag has been deprecated, "
            "use `OPEN_AUTONOMY_PRIVATE_KEY_PASSWORD` to export the password value"
        )

    click.echo(
        "DEPRECATION WARNING: `autonomy deploy from-token` is being deprecated.\n"
        "Please use `autonomy fetch <token_id>` instead to fetch the service, "
        "followed by `build-image`, `deploy build`, and `deploy run` as usual."
    )

    ctx = cast(Context, click_context.obj)
    ctx.registry_type = REGISTRY_REMOTE
    keys_file = Path(keys_file or DEFAULT_KEYS_FILE).absolute()
    with reraise_as_click_exception(
        NotValidKeysFile, FileNotFoundError, FileExistsError
    ):
        build_and_deploy_from_token(
            token_id=token_id,
            keys_file=keys_file,
            chain_type=ChainType(chain_type),
            skip_image=skip_image,
            n=n,
            deployment_type=deployment_type,
            aev=aev,
            no_deploy=no_deploy,
            detach=detach,
            resources={
                "agent": {
                    "limit": {"cpu": agent_cpu_limit, "memory": agent_memory_limit},
                    "requested": {
                        "cpu": agent_cpu_request,
                        "memory": agent_memory_request,
                    },
                }
            },
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the rounds of the skill."""

import json
from typing import Dict
from unittest.mock import MagicMock

from packages.valory.skills.abstract_round_abci.base import (
    BaseSynchronizedData,
    OffenceStatus,
    OffenseStatusEncoder,
)
from packages.valory.skills.abstract_round_abci.test_tools.rounds import (
    BaseCollectSameUntilThresholdRoundTest,
)
from packages.valory.skills.offend_abci.payloads import OffencesPayload
from packages.valory.skills.offend_abci.rounds import Event, OffendRound


class TestOffendRound(BaseCollectSameUntilThresholdRoundTest):
    """Tests for OffendRound."""

    _synchronized_data_class = BaseSynchronizedData
    _event_class = Event

    @property
    def status(
        self,
    ) -> Dict[str, OffenceStatus]:
        """Get a stub offence status."""
        return {participant: OffenceStatus() for participant in self.participants}

    @property
    def serialized_status(
        self,
    ) -> str:
        """Get a stub offence status, serialized."""
        return json.dumps(self.status, cls=OffenseStatusEncoder, sort_keys=True)

    @property
    def participant_to_offend(
        self,
    ) -> Dict[str, OffencesPayload]:
        """Get participant to offend payload mapping."""
        return {
            participant: OffencesPayload(participant, self.serialized_status)
            for participant in self.participants
        }

    def test_run(self) -> None:
        """Runs test."""
        context_mock = MagicMock()
        context_mock.state.round_sequence.offence_status = None
        test_round = OffendRound(self.synchronized_data, context_mock)

        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=self.participant_to_offend,
                synchronized_data_update_fn=lambda _synchronized_data, _: _synchronized_data,
                synchronized_data_attr_checks=[],
                most_voted_payload=self.serialized_status,
                exit_event=Event.DONE,
            )
        )

        assert context_mock.state.round_sequence.offence_status == self.status

    def test_no_majority_event(self) -> None:
        """Test the no-majority event."""
        test_round = OffendRound(self.synchronized_data, MagicMock())
        self._test_no_majority_event(test_round)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the slashing background app composition."""

from packages.valory.skills.abstract_round_abci.abci_app_chain import (
    AbciAppTransitionMapping,
    chain,
)
from packages.valory.skills.slashing_abci.rounds import (
    Event,
    PostSlashingTxAbciApp,
    SlashingCheckRound,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    FailedRound,
    FinishedTransactionSubmissionRound,
    TransactionSubmissionAbciApp,
)


slashing_transition_function: AbciAppTransitionMapping = {
    FinishedTransactionSubmissionRound: PostSlashingTxAbciApp.initial_round_cls,
    FailedRound: TransactionSubmissionAbciApp.initial_round_cls,
}
SlashingAbciApp = chain(
    (
        TransactionSubmissionAbciApp,
        PostSlashingTxAbciApp,
    ),
    slashing_transition_function,
)

SlashingAbciApp.transition_function[SlashingCheckRound] = {
    Event.SLASH_START: TransactionSubmissionAbciApp.initial_round_cls,
}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tendermint/crypto/proof.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from packages.valory.connections.abci.gogoproto import (
    gogo_pb2 as gogoproto_dot_gogo__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1dtendermint/crypto/proof.proto\x12\x11tendermint.crypto\x1a\x14gogoproto/gogo.proto"G\n\x05Proof\x12\r\n\x05total\x18\x01 \x01(\x03\x12\r\n\x05index\x18\x02 \x01(\x03\x12\x11\n\tleaf_hash\x18\x03 \x01(\x0c\x12\r\n\x05\x61unts\x18\x04 \x03(\x0c"?\n\x07ValueOp\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12\'\n\x05proof\x18\x02 \x01(\x0b\x32\x18.tendermint.crypto.Proof"6\n\x08\x44ominoOp\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05input\x18\x02 \x01(\t\x12\x0e\n\x06output\x18\x03 \x01(\t"2\n\x07ProofOp\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c"9\n\x08ProofOps\x12-\n\x03ops\x18\x01 \x03(\x0b\x32\x1a.tendermint.crypto.ProofOpB\x04\xc8\xde\x1f\x00\x42:Z8github.com/tendermint/tendermint/proto/tendermint/cryptob\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "tendermint.crypto.proof_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"Z8github.com/tendermint/tendermint/proto/tendermint/crypto"
    )
    _PROOFOPS.fields_by_name["ops"]._options = None
    _PROOFOPS.fields_by_name["ops"]._serialized_options = b"\310\336\037\000"
    _globals["_PROOF"]._serialized_start = 74
    _globals["_PROOF"]._serialized_end = 145
    _globals["_VALUEOP"]._serialized_start = 147
    _globals["_VALUEOP"]._serialized_end = 210
    _globals["_DOMINOOP"]._serialized_start = 212
    _globals["_DOMINOOP"]._serialized_end = 266
    _globals["_PROOFOP"]._serialized_start = 268
    _globals["_PROOFOP"]._serialized_end = 318
    _globals["_PROOFOPS"]._serialized_start = 320
    _globals["_PROOFOPS"]._serialized_end = 377
# @@protoc_insertion_point(module_scope)
